"""Etl import/export file parser module."""


from typing import (
    List, Optional, Callable, Tuple, Any
)
from dataclasses import replace
from datetime import datetime
import pandas as pd
from tqdm import tqdm

from .model import (
    SeriesSettings,
    SeriesIterator,
    RESOURCE_COLUMN,
    TIMESTAMP_COLUMN,
    METRIC_COLUMN,
    render_timestamp,
    ParserRequestError,
    TIMESTAMP_COLUMN_NAMES,
)


def prepare_settings_dataframe(
    input_data: pd.DataFrame,
    settings: SeriesSettings
) -> SeriesSettings:
    """Validate and update the input settings by extracting metadata from a input Pandas DataFrame."""
    # timestamp specification
    fail_msg = (
        'Dataframe input requires either:\n'
        ' - an index of type DatetimeIndex\n'
        ' - a `timestamp_column` setting indicating a column\n'
        ' - a `timestamp_interval` setting to compute timestamps'
    )
    if isinstance(input_data.index, pd.DatetimeIndex):
        pass
    elif settings.timestamp_interval:
        pass
    else:
        if not settings.timestamp_column:
            for column_name in input_data.columns:
                if isinstance(column_name, str) and column_name.lower() in TIMESTAMP_COLUMN_NAMES:
                    settings = replace(settings, timestamp_column=column_name)
                break
        if settings.timestamp_column:
            if settings.timestamp_column not in input_data.columns:
                raise ParserRequestError(fail_msg)
        else:
            raise ParserRequestError(fail_msg)

    # resource specification
    fail_msg = (
        'Please specify a resource, either as \n'
        ' - a fixed `resource` setting\n'
        ' - a `resource_column` that indicates the column '
        '(or column level) containing the resource'
    )
    if settings.resource:
        if settings.key_by_resource(settings.resource) is None:
            raise ParserRequestError('specified resource not in filtered resources')
    elif settings.resource_column:
        fail_msg = (
            f'Specified invalid `resource_column`={settings.resource_column}.'
            ' This should be either:\n'
            ' - a column name in the input data\n'
            ' - a level in a multiindex column'
        )
        if settings.resource_column in input_data.columns:
            # resources specified in data ...
            pass
        elif isinstance(input_data.columns, pd.MultiIndex):
            if settings.resource_column in input_data.columns.names:
                pass
            else:
                raise ParserRequestError(fail_msg)
        else:
            raise ParserRequestError(fail_msg)
    elif RESOURCE_COLUMN in input_data.columns:
        # resources specified in data ...
        pass
    elif (
        isinstance(input_data.columns, pd.MultiIndex) and
        RESOURCE_COLUMN in input_data.columns.names
    ):
        pass
    else:
        raise ParserRequestError(fail_msg)

    if settings.resources is None:
        settings = replace(
            settings,
            resources=_extract_resources(input_data.columns, settings)
        )

    # metric name specification (column name by default)
    if settings.metric_column:
        fail_msg = (
            f'Specified invalid `metric_column`={settings.metric_column}.'
            ' This should be either:\n'
            ' - a column name in the input data\n'
            ' - a level in a multiindex column'
        )
        if settings.metric_column in input_data.columns:
            pass
        elif isinstance(input_data.columns, pd.MultiIndex):
            if settings.metric_column in input_data.columns.names:
                pass
            else:
                raise ParserRequestError(fail_msg)
        else:
            raise ParserRequestError(fail_msg)

    if settings.metrics is None:
        settings = replace(
            settings,
            metrics=_extract_metrics(input_data.columns, settings)
        )

    # value specification
    if settings.value_column:
        if settings.value_column not in input_data.columns:
            raise ParserRequestError(
                f'Specified invalid `value_column`={settings.value_column}.'
                ' This should be a column name in the input data'
            )
        if not settings.metric_column or METRIC_COLUMN in input_data.columns:
            raise ParserRequestError(
                f'When `value_column`={settings.value_column} is specified, '
                ' a `metric_column` should be specified too.'
            )
    return settings


def iter_timeseries_dataframe(
    input_data: pd.DataFrame,
    settings: SeriesSettings,
    progress: bool = False
) -> SeriesIterator:
    """Create a SeriesIterator from the given pandas Dataframe and settings."""
    settings = prepare_settings_dataframe(input_data, settings)
    resource = settings.resource

    if settings.timestamp_interval:
        if not settings.timestamp_first and not settings.timestamp_last:
            settings.timestamp_last = datetime.utcnow()
        index = pd.date_range(
            start=settings.timestamp_first,
            end=settings.timestamp_last,
            periods=len(input_data.index),
            freq=settings.timestamp_interval,
        )
        index.name = TIMESTAMP_COLUMN
        input_data = input_data.set_index(index)
    else:
        if settings.timestamp_column:
            input_data = input_data.set_index(settings.timestamp_column)
            input_data.index.name = TIMESTAMP_COLUMN

        if settings.timestamp_constructor or not(isinstance(input_data.index, pd.DatetimeIndex)):
            index_series: pd.Series = input_data.index.to_series()
            input_data = input_data.reset_index(drop=True)
            if settings.timestamp_constructor:
                index_series = index_series.apply(settings.timestamp_constructor)
            if not isinstance(index_series, pd.DatetimeIndex):
                index_series = pd.DatetimeIndex(index_series)
            input_data.index = index_series

        offsets = [
            offset for offset in [
                settings.timestamp_offset,
                settings.timestamp_first - input_data.index[0]
                if settings.timestamp_first else None,
                settings.timestamp_last - input_data.index[-1]
                if settings.timestamp_last else None
            ]
            if offset
        ]
        if len(offsets) > 1:
            raise ParserRequestError(
                'To offset the timestamps for an ETL input, please specify only one of \n'
                ' - `timestamp_offset`\n'
                ' - `timestamp_first`\n'
                ' - `timestamp_last`\n'
            )
        if offsets:
            input_data = input_data.set_index(input_data.index + offsets[0])

        # assert correct tz info:
        if input_data.index.size > 0:
            first_ts = input_data.index[0]
            render_first_ts = render_timestamp(first_ts)

            if render_first_ts.endswith('Z'):
                pass
            elif '+' in render_first_ts:
                input_data = input_data.set_index(input_data.index.tz_convert('UTC'))
            elif settings.timestamp_timezone:
                index_series = input_data.index.tz_localize(settings.timestamp_timezone)
                index_series = index_series.tz_convert('UTC')
                input_data = input_data.set_index(index_series)
            else:
                raise ParserRequestError(
                    'Input data should contain (or have rules to convert to) UTC timestamps (including timezone info)\n'
                    f'  First timestamp: {first_ts}, rendered as {render_timestamp(first_ts)}\n'
                    'Please provide either:\n'
                    ' - a `timestamp_timezone` to interpret (local) timestamps to an UTC, e.g.\n'
                    '         timestamp_timezone="Europe/Brussels"\n'
                    ' - a `timestamp_constructor` to convert timestamps to an UTC timestamp, e.g.\n'
                    '         timestamp_constructor=lambda t: pd.Timestamp(t, tz="Europe/Brussels")\n'
                    ' - a `timestamp_first` or `timestamp_last` to shift the timestamps'
                )

    key_lambda = _column_key_extractor(input_data.columns, settings)
    if key_lambda is None:
        raise ParserRequestError('NOT YET SUPPORTED: variable resource and metric by row')
    if settings.value_column:
        column_series_iter = iter([
            (settings.value_column, input_data[settings.value_column])
        ])
        series_count = 1
    else:
        column_series_iter = input_data.iteritems()
        series_count = len(input_data.columns)
    if progress and series_count > 1:
        with tqdm(total=series_count, unit_scale=True, unit_divisor=1, unit="series") as tqdm_progress:
            for column, series in column_series_iter:
                resource_column, metric_key = key_lambda(column)
                metric = settings.metric_by_key(metric_key)
                resource = settings.resource_by_key(resource_column)
                if resource and metric:
                    yield resource, metric, ((t, v) for t, v in series.items())
                    tqdm_progress.update(1)
    else:
        for column, series in column_series_iter:
            resource_column, metric_key = key_lambda(column)
            metric = settings.metric_by_key(metric_key)
            resource = settings.resource_by_key(resource_column)
            if resource and metric:
                yield resource, metric, ((t, v) for t, v in series.items())


def _column_key_extractor(columns, settings: SeriesSettings) -> Optional[Callable[[Any], Tuple[str, str]]]:
    default_resource = settings.resource
    if not isinstance(columns, pd.MultiIndex):
        if default_resource is None:
            if (
                settings.resource_column in columns or
                RESOURCE_COLUMN in columns
            ):
                return None  # different resource by row
            raise ParserRequestError('no resource specified')
        default_resource_id: str = default_resource
        if settings.metric_column in columns:
            return None  # different resource by row
        return lambda col: (default_resource_id, col)

    # multi index level containing the resource id
    resource_idx: int = next(
        (
            columns.names.index(resource_column)
            for resource_column in [
                settings.resource_column,
                RESOURCE_COLUMN
            ]
            if resource_column in columns.names
        ),
        -1
    )

    # multi index level containing the metric name
    metric_key_idx: int = next(
        (
            columns.names.index(resource_column)
            for resource_column in [
                settings.metric_column,
                METRIC_COLUMN,
                'name'
            ]
            if resource_column in columns.names
        ),
        -1
    )

    if resource_idx == -1:
        if default_resource is None:
            raise ParserRequestError('no resource specified')

        default_resource_str: str = default_resource
        if metric_key_idx == -1:
            # use all levels joined as metric name
            return lambda col: (default_resource_str, '-'.join(col))

        return lambda col: (default_resource_str, col[metric_key_idx])
    else:
        if metric_key_idx == -1:
            # use all levels joined as metric name (except the resource one)
            return lambda col: (col[resource_idx], '-'.join(col[:resource_idx]+col[(resource_idx+1):]))

        return lambda col: (col[resource_idx], col[metric_key_idx])


def _extract_resources(columns, settings) -> Optional[List[str]]:
    extractor = _column_key_extractor(columns, settings)
    if not extractor:
        return None

    return list({
        extractor(col)[0]: True
        for col in columns
        if col != settings.timestamp_column
    })


def _extract_metrics(columns, settings) -> Optional[List[str]]:
    extractor = _column_key_extractor(columns, settings)
    if not extractor:
        return None

    return list({
        extractor(col)[1]: True
        for col in columns
        if col != settings.timestamp_column
    })



"""Resource action method decorators specific for the 'byoml' service."""

from functools import wraps
from tenacity import (
    Retrying,
    before_sleep,
    stop_after_attempt,
    stop_after_delay,
    wait_exponential,
    before_sleep_log,
    retry_if_exception
)
import logging

from simple_rest_client.exceptions import ErrorWithResponse

from ._exceptions import (
    ByomlActionError
)

logger = logging.getLogger(__name__)

TEMPORARY_FAILURES = [
    409,
    429,
    500,
    502,
    503,
    504,
    508
]
WAIT_EXPONENTIAL = {
    "multiplier": 1,
    "min": 4,
    "max": 32
}
DEFAULT_RETRY_ATTEMPTS = 20
DEFAULT_RETRY_MAX_DELAY = 120


def byoml_exception_decorator(action_method):
    """Create a decorator that parses json error responses."""
    @wraps(action_method)
    def wrapped(*args, **kwargs):
        try:
            return action_method(*args, **kwargs)
        except ErrorWithResponse as exc:
            raise ByomlActionError.from_cause(exc) from exc
    return wrapped


def byoml_retry_decorator(action_method):
    """Create a decorator that retries after certain exceptions."""
    @wraps(action_method)
    def wrapped(*args, **kwargs):
        retry_attempts = kwargs.pop('retry_attempts', DEFAULT_RETRY_ATTEMPTS)
        retry_max_delay = kwargs.pop('retry_max_delay', DEFAULT_RETRY_MAX_DELAY)
        for attempt in Retrying(
            stop=(stop_after_attempt(retry_attempts) | stop_after_delay(retry_max_delay)),
            retry=retry_if_exception(is_temporary_failure),
            before_sleep=before_sleep_log(logger, logging.INFO),
            wait=wait_exponential(**WAIT_EXPONENTIAL),
            reraise=True
        ):
            with attempt:
                return action_method(*args, **kwargs)
    retry_attempts_arg = f'retry_attempts={DEFAULT_RETRY_ATTEMPTS}'
    retry_max_delay_arg = f'retry_max_delay={DEFAULT_RETRY_MAX_DELAY}'
    wrapped.__doc__ = ((wrapped.__doc__ or '') + f"""
Retries on temporary failures (HTTP status codes {', '.join(str(c) for c in TEMPORARY_FAILURES)}).
Arguments that configure retry:
    {retry_attempts_arg:30}: Maximal number of retries.
    {retry_max_delay_arg:30}: Maximal delay in seconds.
""")
    return wrapped


def is_temporary_failure(exc):
    """Check if given exception is temporary."""
    return hasattr(exc, "response") and exc.response.status_code in TEMPORARY_FAILURES



"""Exceptions specific to the Byoml Service."""

from ...exceptions import RestResponseError, RestResponseParseError, RequestError


class ByomlActionError(RestResponseError):
    """Error that represents the json messages of a byoml response."""

    @property
    def message(self):
        """Get the main user error returned by an analytics error response."""
        return self._get_from_body('error', super().message)


class ByomlActionParseError(RestResponseParseError, ByomlActionError):
    """Indicates that a byoml response could not be parsed."""


class ByomlValidationError(RequestError):
    """Exception class for BYOML validation errors."""



"""REST definitions for the 'series' entity of the 'data' service."""

from waylay.service import WaylayResource
from waylay.service import decorators

DEFAULT_DECORATORS = [decorators.exception_decorator, decorators.return_body_decorator]


class EventsResource(WaylayResource):
    """REST Resource for the 'events' ingestion of the 'data' service."""

    link_roots = {
        'doc': '${doc_url}/api/broker-and-storage/#',
        'iodoc': '${iodoc_url}/api/broker/?id='
    }

    actions = {
        'post': {
            'method': 'POST', 'url': '/resources/{}/events',
            'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator([]),
            ],
            'description': (
                'Forward a json message to the rule engine, '
                'time series database and/or document store for a given resource.'
            ),
            'links': {
                'doc': 'posting-data-object-to-the-storage-and-rule-engine',
                'iodoc': 'posting-data-to-the-storage-and-rule-engine'
            },
        },
        'bulk': {
            'method': 'POST', 'url': '/messages',
            'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator([])
            ],
            'description': (
                'Forward an array of json messages to the rule engine, '
                'time series database and/or document store.'
            ),
            'links': {
                'doc': 'posting-array-of-objects',
                'iodoc': 'posting-array-of-data'
            },
        },
        'remove': {
            'method': 'DELETE', 'url': '/resources/{}',
            'decorators': DEFAULT_DECORATORS,
            'description': 'Remove all data for a resource.',
            'links': {
                'doc': 'all-data-for-a-resource',
                'iodoc': 'all-data-for-a-resource'
            },
        },
    }



"""REST definitions for the 'series' entity of the 'data' service."""

from waylay.service import WaylayResource
from waylay.service import decorators

DEFAULT_DECORATORS = [decorators.exception_decorator, decorators.return_body_decorator]


class SeriesResource(WaylayResource):
    """REST Resource for the 'series' entity of the 'data' service."""

    link_roots = {
        'doc': '${doc_url}/api/broker-and-storage/#',
        'iodoc': '${iodoc_url}/api/broker/?id='
    }

    actions = {
        'data': {
            'method': 'GET', 'url': '/resources/{}/series/{}',
            'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator(['series']),
            ],
            'description': 'Retrieve the (optionally aggregated) data of a single series.',
            'links': {
                'doc': 'getting-time-series-data',
                'iodoc': 'getting-time-series-data'
            },
        },
        'list': {
            'method': 'GET', 'url': '/resources/{}/series',
            'decorators': [
                decorators.default_params_decorator({'metadata': 'true'}),
                decorators.exception_decorator,
                decorators.return_path_decorator([]),
            ],
            'description': 'Retrieve a list of series and their latest value for a given resource.',
            'links': {
                'doc': 'metadata',
                'iodoc': 'metadata'
            },
        },
        'latest': {
            'method': 'GET', 'url': '/resources/{}/series/{}/latest',
            'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator([])
            ],
            'description': 'Fetch the latest value for a series.',
            'links': {
                'doc': 'latest-value-for-a-series',
                'iodoc': 'latest-value-for-a-series'
            },
        },
        'query': {
            'method': 'POST', 'url': '/series/query',
            'description': 'Execute a broker query document to retrieve aggregated timeseries.',
            'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator(['series'])
            ],
            'links': {
                'doc': 'post-timeseries-query',
                'iodoc': 'post-timeseries-query'
            },
        }
    }



"""Command line interface for the Waylay Python SDK."""

import argparse
import logging
import sys

from waylay import WaylayClient


def init_doc_parser(parser):
    """Initialize the parser for the `doc` command."""
    parser.add_argument(
        '-s', '--service', dest='doc_service', nargs='*',
        help='Filter services to document.', default=None
    )
    parser.add_argument(
        '-r', '--resource', dest='doc_resource', nargs='*',
        help='Filter resources to document.', default=None
    )
    parser.add_argument(
        '-l', '--link', dest='doc_link', nargs='*', help='Filter doc sites links.', default=None
    )


def init_srv_parser(parser):
    """Initialize the parser for the `service` command."""
    cmd_parsers = parser.add_subparsers(dest='srv_cmd')
    cmd_parsers.add_parser('list', help='List services.')


def main():
    """Start the waylaycli program."""
    logging.basicConfig()
    parser = argparse.ArgumentParser(
        prog='waylaycli', description='Command line interface to the Waylay Python SDK'
    )
    parser.add_argument(
        '-p', '--profile', dest='profile', nargs='?', help='Configuration profile.', default=None
    )
    cmd_parsers = parser.add_subparsers(dest='cmd')
    init_doc_parser(cmd_parsers.add_parser('doc', help='Generate SDK overview of services.'))
    init_srv_parser(cmd_parsers.add_parser('service', help='Services.'))

    args = parser.parse_args()

    def waylay_client():
        return WaylayClient.from_profile(args.profile)

    if args.cmd == 'doc':
        print(waylay_client().util.info.action_info_html(
            service=args.doc_service, resource=args.doc_resource, links=args.doc_link
        ))
        return
    if args.cmd == 'service' and args.srv_cmd == 'list':
        profile = f'profile "{args.profile}"' if args.profile else 'default profile'
        print(
            f'{"key":^10} '
            f'| {" url for "+profile:^35} '
            f'| description'
        )
        print('-' * 10 + '   ' + '-' * 35 + '   ' + '-' * 30)
        for srv in waylay_client().services:
            print(
                f'{srv.service_key:>10} '
                f'| {srv.root_url:<35} '
                f'| {srv.description}'
            )
        return
    parser.print_help()



"""REST definitions for version status of Waylay Storage service."""
from urllib.parse import quote
from waylay.service import WaylayResource, decorators

DEFAULT_DECORATORS = [decorators.exception_decorator, decorators.return_body_decorator]


class AboutResource(WaylayResource):
    """Static information about version."""

    link_roots = {
        'iodoc': '${iodoc_url}/api/storage/?id=',
        'apidoc':  '${root_url}/docs#/status/',
        'openapi': '${root_url}/openapi.json'
    }

    actions = {
        'version': {
            'method': 'GET',
            'url': '/',
            'decorators': DEFAULT_DECORATORS,
            'description': 'Application version',
            'links': {
                'iodoc': 'version',
                'apidoc': 'version__get',
                'openapi': f"#/paths/{quote('/')}/get",
            },
        },
        'status': {
            'method': 'GET',
            'url': '/status',
            'decorators': DEFAULT_DECORATORS,
            'description': 'Validation and statistics on the buckets and policies for this tenant.',
            'links': {
                'iodoc': 'tenant-status',
                'apidoc': 'status_status_get',
                'openapi': f"#/paths/{quote('/status')}/get",
            },
        },
    }



"""Python SDK Documentation tooling."""

from typing import Optional, Dict, Union, List

import pandas as pd
import numpy as np

from waylay.service import WaylayServiceContext


def format_links_html(links: Optional[Dict[str, Dict[str, str]]]) -> str:
    """Format link information to an html string."""
    if not links:
        return ''
    link_items = (
        f'<a href="{href}" target="{target}">{text}</a>'
        for target, href, text in (
            (f"_{key}", link.get('href'), link.get('title', key))
            for key, link in links.items()
        )
        if href
    )
    if not link_items:
        return ''
    return " | ".join(link_items)


class InfoTool:
    """Client tool that exposes information about the provided services, recources and operations."""

    def __init__(self, context: WaylayServiceContext):
        """Create a _info_ tool."""
        self._service_context = context

    def action_info_df(
        self,
        service: Optional[Union[str, List[str]]] = None,
        resource: Optional[Union[str, List[str]]] = None,
        links: Optional[Union[str, List[str]]] = None,
    ) -> pd.DataFrame:
        """Produce a pandas DataFrame with an overview of the provided actions.

        Parameters:
            service     filter on the services to be listed
            resource    filter on the resources to be listed
            links       filter on the documentation link names
        """
        def service_filter(name):
            return service is None or name == service or name in service

        def resource_filter(name):
            return resource is None or name == resource or name in resource

        def format_filtered_links(hal_links):
            if not hal_links:
                return ''
            return format_links_html(
                {
                    link: ref
                    for link, ref in hal_links.items()
                    if not links or link in links
                }
            )

        def format_pydoc(docstr: Optional[str]) -> Optional[str]:
            if not docstr:
                return docstr
            return f'<pre>{docstr.strip()}</pre>'

        def action_doc(action) -> str:
            return "\n".join(f'<div>{doc}</div>' for doc in [
                action.description,
                format_pydoc(action.sdk_action_doc),
                format_filtered_links(action.hal_links)
            ] if doc)

        df_doc = pd.DataFrame(np.transpose([
            [
                service.service_key,
                resource.resource_name,
                action.name,
                action.method,
                action.url,
                action_doc(action),
            ]
            for service in self._service_context.list()
            if service_filter(service.name)
            for resource in service.list_resources()
            if resource_filter(resource.name)
            for action_name, action in resource.actions.items()
        ]), index=[
            'service',
            'resource',
            'action',
            'method',
            'url',
            'description'
        ], ).T
        return df_doc.set_index(['service', 'resource', 'action'])

    def action_info_html(
        self,
        service: Optional[Union[str, List[str]]] = None,
        resource: Optional[Union[str, List[str]]] = None,
        links: Optional[Union[str, List[str]]] = None,
    ) -> str:
        """Render the service/resource/action listing as an html table.

        Parameters:
            service     filter on the services to be listed
            resource    filter on the resources to be listed
            links       filter on the documentation link names
        """
        html = self.action_info_df(service, resource, links).to_html(
            escape=False,
            header=False,
            index_names=False,
        )
        # custom header
        html = html.replace('<tbody>', """<thead>
  <tr>
    <th>service</th>
    <th>resource</th>
    <th>action</th>
    <th>method</th>
    <th>url</th>
    <th>description</th>
  </tr>
</thead>
<tbody>""")
        # replace double escapes in html
        return html.replace('\\n', "\n")



"""Conversion from CSV to elt export format."""

from typing import (
    Union, Sequence, Iterator, Iterable,
    Optional, Tuple,
    cast
)
from dataclasses import replace
import os
import io
import csv
import gzip
from contextlib import contextmanager

from tqdm import tqdm

from .model import (
    SeriesInput,
    SeriesSettings,
    SeriesIterator,
    MeasurementIterator,
    Metric,
    CSVReader,
    PathLike,
    METRIC_PREFIX,
    METRIC_PREFIX_LENGTH,
    ETL_IMPORT_COLUMN_NAMES,
    RESOURCE_COLUMN,
    METRIC_COLUMN,
    VALUE_COLUMN,
    ParserRequestError,
    default_create_timestamp,
    TIMESTAMP_COLUMN_NAMES,
)


def prepare_settings_csv(
    input_data: SeriesInput,
    settings: SeriesSettings,
) -> SeriesSettings:
    """Validate and update the input settings by extracting metadata from a csv header."""
    with open_csv(input_data) as csv_data:
        header = next(csv_data)

        settings = _prepare_csvheader_timestamp(header, settings)

        settings = _prepare_csvheader_resource_metric(header, settings)

        if settings.value_column or all(column in header for column in ETL_IMPORT_COLUMN_NAMES):
            # single column contains values, metric must be provided fixed or as other column.
            settings = _prepare_csvheader_value_column(header, settings)
        else:
            # series specified in seperate columns
            settings = _prepare_csvheader_series_column(header, settings)

    return settings


def _prepare_csvheader_timestamp(header: Sequence[str], settings: SeriesSettings) -> SeriesSettings:
    # time column handling
    if settings.timestamp_column:
        if settings.timestamp_column not in header:
            raise ParserRequestError(
                f'Timestamp column `{settings.timestamp_column}` not in csv header.'
            )
    elif settings.timestamp_first and settings.timestamp_interval:
        raise NotImplementedError('timestamp_first timestamp_interval setting not supported')
    else:
        for timestamp_column in TIMESTAMP_COLUMN_NAMES:
            if timestamp_column in header:
                settings = replace(settings, timestamp_column=timestamp_column)
                break
        if not settings.timestamp_column:
            raise ParserRequestError(
                f'No timestamp column found in `{header}`'
            )

    # default timestamp value constructor
    if settings.timestamp_column:
        timestamp_constructor = settings.timestamp_constructor or default_create_timestamp
        if settings.timestamp_timezone:
            if settings.timestamp_constructor:
                raise ParserRequestError(
                    'Cannot specify both `timestamp_constructor` and `timestamp_timezone`.'
                )
            timezone = settings.timestamp_timezone

            def timestamp_constructor_localized(ts):
                return default_create_timestamp(ts).tz_localize(timezone)
            timestamp_constructor = timestamp_constructor_localized
            settings = replace(
                settings,
                timestamp_timezone=None
            )

        if settings.timestamp_offset:
            offset = settings.timestamp_offset
            original_timestamp_constructor = timestamp_constructor

            def timestamp_constructor_offset(ts):
                return original_timestamp_constructor(ts) + offset

            timestamp_constructor = timestamp_constructor_offset

        settings = replace(
            settings,
            timestamp_constructor=timestamp_constructor
        )
    return settings


def _prepare_csvheader_resource_metric(header: Sequence[str], settings: SeriesSettings) -> SeriesSettings:
    if not settings.resource or settings.resource_column:
        settings = replace(settings, resource_column=settings.resource_column or RESOURCE_COLUMN)
        if settings.resource_column not in header:
            raise ParserRequestError(
                f'No valid resource or resource key provided: `{settings.resource_column}`'
            )
    if settings.metric_column:
        if settings.metric_column not in header:
            raise ParserRequestError(
                f'Invalid metric key provided: `{settings.metric_column}`'
            )
    return settings


def _prepare_csvheader_value_column(header: Sequence[str], settings: SeriesSettings) -> SeriesSettings:
    if not settings.value_column:
        settings = replace(settings, value_column=VALUE_COLUMN)
    if settings.value_column not in header:
        raise ParserRequestError(
            f'Value column `{settings.value_column}` not in csv header.'
        )
    if not settings.metric and not settings.metric_column:
        if METRIC_COLUMN in header:
            settings = replace(settings, metric_column=METRIC_COLUMN)
        else:
            raise ParserRequestError(
                f'No valid metric or metric key provided for values in `{settings.value_column}` column.'
            )
    return settings


def _prepare_csvheader_series_column(header: Sequence[str], settings: SeriesSettings) -> SeriesSettings:
    specified_metrics = list(settings.iter_metrics())
    if not specified_metrics:
        specified_metrics = [
            Metric(name=column)
            for column in header
            if column != settings.timestamp_column
            if column != settings.resource_column
        ]

    specified_metrics = [
        metric_spec
        for metric_spec in specified_metrics
        if metric_spec.key_or_name in header
    ]
    if not specified_metrics:
        raise ParserRequestError(
            f'None of the specified metrics `{settings.metrics}` found in the csv header.'
        )
    return replace(settings, metrics=specified_metrics)


def iter_timeseries_csv(
    series_input: SeriesInput,
    settings: SeriesSettings,
    progress: bool = False
) -> SeriesIterator:
    """Create a SeriesIterator from the given csv input and settings.

    This reads the csv input at least once for each series.
    """
    settings = prepare_settings_csv(series_input, settings)
    resource_metric_keys = _resource_metric_keys_csv(series_input, settings)
    if progress:
        with tqdm(total=len(resource_metric_keys), unit_scale=True, unit_divisor=1, unit="series") as tqdm_progress:
            for resource_key, metric_key in resource_metric_keys:
                yield _get_timeseries_csv(series_input, settings, resource_key, metric_key)
                tqdm_progress.update(1)
    else:
        for resource_key, metric_key in resource_metric_keys:
            yield _get_timeseries_csv(series_input, settings, resource_key, metric_key)


def _get_timeseries_csv(
    series_input: SeriesInput,
    settings: SeriesSettings,
    resource_key: Optional[str],
    metric_key: str
) -> Tuple[str, str, MeasurementIterator]:
    resource = settings.resource_by_key(resource_key)
    assert resource
    metric = settings.metric_by_key(metric_key)
    assert metric

    if settings.value_column:
        # fixed value column
        return (resource, metric, _iter_timeseries_csv_one_series(
            series_input,
            settings,
            value_column=settings.value_column,
            resource_key_filter=resource_key,
            metric_key_filter=metric_key
        ))

    # fixed column: each one is a series
    return (resource, metric, _iter_timeseries_csv_one_series(
        series_input,
        settings,
        value_column=metric_key,
        resource_key_filter=resource_key,
    ))


@contextmanager
def open_csv(series_input: SeriesInput, **csv_format_args) -> Iterator[CSVReader]:
    """Open the CSV file specified by this input."""
    if isinstance(series_input, (str, os.PathLike)):
        if str(series_input).endswith('.gz'):
            yield from _open_csv_gz_file(series_input, **csv_format_args)

        else:
            yield from _open_csv_file(series_input, **csv_format_args)

    elif isinstance(series_input, io.TextIOBase):
        yield from _reset_and_open_csv_text_stream(series_input, **csv_format_args)

    elif isinstance(series_input, Iterable):
        yield from _open_csv_iterable(series_input, **csv_format_args)


def _open_csv_gz_file(series_input: PathLike, **csv_format_args) -> Iterator[CSVReader]:
    with gzip.open(series_input, 'rt') as csv_file:
        yield csv.reader(cast(io.TextIOBase, csv_file), **csv_format_args)


def _open_csv_file(series_input: PathLike, **csv_format_args) -> Iterator[CSVReader]:
    with open(series_input, 'rt') as csv_file:
        yield csv.reader(csv_file, **csv_format_args)


def _reset_and_open_csv_text_stream(series_input: io.TextIOBase, **csv_format_args) -> Iterator[CSVReader]:
    series_input.seek(0)
    yield csv.reader(series_input, **csv_format_args)


def _open_csv_iterable(
    series_input: Union[Iterable[str], Iterable[Sequence[str]]], **csv_format_args
) -> Iterator[CSVReader]:
    # inspect first item of iterable
    first_line = next(iter(series_input))
    if isinstance(first_line, str):
        # Iterable of strings, treat as unparsed csv data
        yield csv.reader(
            iter(cast(Iterable[str], series_input)),
            **csv_format_args
        )
    elif isinstance(first_line, Sequence):
        # Iterable of lists, treat as parsed csv data
        yield iter(cast(Iterable[Sequence[str]], series_input))
    else:
        raise ParserRequestError(
            f'Cannot read first line of a csv input:\n{first_line}'
        )
    return


def _resource_metric_keys_csv(
    series_input: SeriesInput,
    settings: SeriesSettings,
) -> Sequence[Tuple[Optional[str], str]]:
    fixed_metric_keys = [
        metric_spec.key_or_name for metric_spec in settings.iter_metrics()
    ]
    fixed_resource_keys = [
        resource_spec.key_or_id for resource_spec in settings.iter_resources()
    ]
    if not settings.resource_column and not settings.metric_column:
        assert settings.resource
        resource_key = settings.key_by_resource(settings.resource)
        assert resource_key
        return [
            (resource_key, metric_key) for metric_key in fixed_metric_keys
        ]

    # use an ordered dict rather than a set to retain insertion order
    resource_metric_keys_set = dict(
        (item, True)
        for item in _iter_resource_metric_keys_csv(series_input, settings)
    )
    return [
        (resource_key, metric_key)
        # (resource_key , metric_key_val) as defined in the data set
        for resource_key, metric_key_val in resource_metric_keys_set.keys()
        # if metric_key_val is not defined in data itself,
        # use the fixed metric keys (from column headers, or as specified by user)
        for metric_key in (
            [metric_key_val] if metric_key_val else fixed_metric_keys
        )
        # filter the metric_key (from data) if a filtering was given
        if not fixed_metric_keys or _remove_prefix(metric_key) in fixed_metric_keys
        # filter the resource_key (from data) if a filtering was given
        if not resource_key or not fixed_resource_keys or resource_key in fixed_resource_keys
    ]


def _remove_prefix(metric_key: str) -> str:
    if metric_key.startswith(METRIC_PREFIX):
        return metric_key[METRIC_PREFIX_LENGTH:]
    return metric_key


def _iter_resource_metric_keys_csv(
    series_input: SeriesInput,
    settings: SeriesSettings,
) -> Iterator[Tuple[Optional[str], Optional[str]]]:
    # either the resource or metric is specified as a column,
    # loop once over the complete data set to extract the combinations.
    # A None in the resulting tuple indicates that the
    # resource or metric is not defined as data, but
    # as default or row header.

    with open_csv(series_input) as csv_iter:
        header = next(csv_iter)
        resource_idx = None
        if settings.resource_column:
            resource_idx = header.index(settings.resource_column)

        metric_idx = None
        if settings.metric_column:
            metric_idx = header.index(settings.metric_column)

        for item in csv_iter:
            if not item:  # ignore empty lines
                continue
            yield (
                item[resource_idx] if resource_idx is not None else None,
                item[metric_idx] if metric_idx is not None else None
            )


def _iter_timeseries_csv_one_series(
    series_input: SeriesInput,
    settings: SeriesSettings,
    value_column: str,
    resource_key_filter: Optional[str] = None,
    metric_key_filter: Optional[str] = None,
) -> MeasurementIterator:
    with open_csv(series_input) as csv_iter:
        # inspect header
        header = next(csv_iter)
        timestamp_idx = header.index(settings.timestamp_column)
        value_idx = header.index(value_column)
        create_timestamp = settings.timestamp_constructor
        assert create_timestamp

        # Create yield loops (avoid unnessesary checks in tight loops).
        # This is a single loop that is unrolled into four cases,
        # to minimize the number of condition checks within the loop.
        resource_idx = None
        metric_idx = None
        if resource_key_filter and settings.resource_column:
            resource_idx = header.index(settings.resource_column)
        if metric_key_filter and settings.metric_column:
            metric_idx = header.index(settings.metric_column)

        if resource_idx is not None and metric_idx is not None:
            # Case 1: filtering on resource AND metric columns
            for item in csv_iter:
                if (
                    item
                    and item[metric_idx] == metric_key_filter
                    and item[resource_idx] == resource_key_filter
                ):
                    yield (create_timestamp(item[timestamp_idx]), item[value_idx])
            return

        if resource_idx is not None:
            # Case 2: filtering on resource column only
            for item in csv_iter:
                if item and item[resource_idx] == resource_key_filter:
                    yield (create_timestamp(item[timestamp_idx]), item[value_idx])
            return

        if metric_idx is not None:
            # Case 3: filtering on metric column only
            for item in csv_iter:
                if item and item[metric_idx] == metric_key_filter:
                    yield (create_timestamp(item[timestamp_idx]), item[value_idx])
            return

        # Case 4: no filtering (each row contains data for the series)
        for item in csv_iter:
            if item:
                yield (create_timestamp(item[timestamp_idx]), item[value_idx])
        return



"""Waylay Python SDK."""

from .client import WaylayClient
from .config import WaylayConfig
from .auth import CredentialsType, ClientCredentials
from .exceptions import WaylayError, RestResponseError
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



"""REST client for the Waylay ETL Service."""

from waylay.service import WaylayService

from .import_ import ImportResource


class ETLService(WaylayService):
    """REST client for the Waylay ETL Service."""

    config_key = 'etl'
    service_key = 'etl'
    default_root_url = 'https://etl.waylay.io'
    resource_definitions = {
        'etl_import': ImportResource,
    }

    etl_import: ImportResource




"""REST client for the Waylay Data Service (broker)."""

from waylay.service import WaylayService

from .series import SeriesResource
from .events import EventsResource


class DataService(WaylayService):
    """REST client for the Waylay Data Service (broker)."""

    config_key = 'data'
    service_key = 'data'
    default_root_url = 'https://data.waylay.io'
    resource_definitions = {
        'series': SeriesResource,
        'events': EventsResource,
    }

    series: SeriesResource
    events: EventsResource



"""resource action method decorators specific for the 'analytics' service."""

from functools import wraps
from enum import Enum
from typing import (
    Union, List
)
import pandas as pd
from simple_rest_client.exceptions import ErrorWithResponse

from ._util import (
    parse_df,
    RENDER_MODE_COMPACT_WS
)
from ._exceptions import (
    AnalyticsActionError,
    AnalyticsActionParseError
)

DEFAULT_RENDER_MODE = RENDER_MODE_COMPACT_WS


def analytics_exception_decorator(action_method):
    """Decorate an analytics action with an error response parser."""
    @wraps(action_method)
    def wrapped(*args, **kwargs):
        try:
            return action_method(*args, **kwargs)
        except ErrorWithResponse as exc:
            raise AnalyticsActionError.from_cause(exc) from exc
    return wrapped


class MultiFrameHandling(Enum):
    """Strategies to handle list of data frames as returned by the the analytics api."""

    FIRST = 'Return the first dataframe'
    LIST = 'Return a list of dataframes'
    SINGLE = 'Assert that the response contains a single dataframe'
    SINGLE_OR_LIST = 'Return a dataframe or a list of dataframes'
    JOIN = 'Join all dataframes, using table attributes'


def analytics_return_dataframe_decorator(
    data_key: str = 'data',
    default_render_mode: str = DEFAULT_RENDER_MODE,
    default_frames_handling: MultiFrameHandling = MultiFrameHandling.SINGLE
):
    """Decorate an analytics action with dataframe representation handling."""

    def decorator(action_method):
        @wraps(action_method)
        def wrapped(*args, **kwargs):
            raw = kwargs.pop('raw', False)
            constructor = kwargs.pop('response_constructor', 'tsa_dataframe') or (lambda x: x)
            frames_handling = kwargs.pop('multi_frame', default_frames_handling)
            # 'render_mode' call argument has precendence on 'render.mode' url param
            render_mode = kwargs.pop('render_mode', None)
            if not render_mode and 'params' in kwargs:
                render_mode = kwargs['params'].get('render.mode', None)
            if 'params' not in kwargs:
                kwargs['params'] = {}
            render_mode = render_mode or default_render_mode
            # set render mode
            kwargs['params']['render.mode'] = render_mode
            try:
                resp = action_method(*args, **kwargs)
            except ErrorWithResponse as exc:
                raise AnalyticsActionError.from_cause(exc) from exc
            if raw:
                return resp
            try:
                if constructor == 'tsa_dataframe':
                    return parse_data_frame_response(
                        resp.body.get(data_key), render_mode, frames_handling
                    )
                return constructor(resp.body.get(data_key))
            except (ValueError, AttributeError, TypeError) as exc:
                raise AnalyticsActionParseError(exc.args[0], resp) from exc
        wrapped.__doc__ = ((wrapped.__doc__ or '') + f"""
Returns a pandas dataframe. Arguments that configure this behaviour:
    `render_mode={DEFAULT_RENDER_MODE}`: render mode requested to the analytics server.
    `response_constructor='tsa_dataframe'`: converter for the data response.
    `multi_frame=MultiFrameHandling.{default_frames_handling.name}`: {default_frames_handling.value}.
""")
        return wrapped
    return decorator


def parse_data_frame_response(
    data_response, render_mode, frames_handling
) -> Union[None, pd.DataFrame, List[pd.DataFrame]]:
    """Parse a json reponse to a dataframe.

    Handles a list of dataframes as specified in `frames_handling`.
    """
    if isinstance(data_response, list):
        if len(data_response) == 0:
            return None

        has_multiple = len(data_response) > 1

        if frames_handling == MultiFrameHandling.SINGLE and has_multiple:
            raise ValueError(
                'multiple dataframes returned while one expected'
            )

        df_list = [parse_df(resp_df, render_mode=render_mode) for resp_df in data_response]

        if frames_handling == MultiFrameHandling.LIST:
            return df_list

        if not has_multiple:
            return df_list[0]

        if frames_handling == MultiFrameHandling.FIRST:
            return df_list[0]

        if frames_handling == MultiFrameHandling.JOIN:
            # TODO
            # * unroll table attributes into column attributes
            # * join tables
            # * move 'single level' attributes back to table attributes (except for levels)
            #   that were in the columns originally
            # what follows is a hack that for our common 'role' case works
            return pd.concat({
                df.tsa.table_attributes['role']: df for df in df_list
            }, axis=1, names=['role'])

        if frames_handling == MultiFrameHandling.SINGLE_OR_LIST:
            return df_list
        raise ValueError(
            f'could not handle multipe dataframes with {frames_handling}'
        )
    else:
        return parse_df(data_response)



"""REST definitions for the import process entity of the etl service."""

from waylay.service import WaylayResource
from waylay.service import decorators

DEFAULT_DECORATORS = [decorators.exception_decorator, decorators.return_body_decorator]


class ImportResource(WaylayResource):
    """REST Resource for the import process entity of the etl service."""

    link_roots = {
        'iodoc': '${iodoc_url}/features/etl/?id='
    }

    actions = {
        'initiate': {
            'method': 'POST',
            'url': '/api/etl/import',
            'decorators': DEFAULT_DECORATORS,
            'description': "Initiates an etl import process as specified in the request body.",
            'links': {
                'iodoc': 'etl-import-service',
            }
        },
        'get': {
            'method': 'GET',
            'url': '/api/etl/import',
            'decorators': DEFAULT_DECORATORS,
            'description': "Retrieves the last or active etl import process.",
            'links': {
                'iodoc': 'etl-import-service',
            }
        }
    }



"""Plugin Package for the Timeseries Service tooling."""

from ._service import TimeSeriesService


from .parser.model import (
    Resource, Metric, SeriesSettings, ETLFile, WaylayETLSeriesImport,
    ParserRequestError
)




"""REST client for the Waylay Time Series Analytics Service."""

from .._base import WaylayService

from .query import QueryResource
from .about import AboutResource


class AnalyticsService(WaylayService):
    """REST client for the Analytics Service."""

    service_key = 'analytics'
    config_key = 'analytics'
    default_root_url = 'https://ts-analytics.waylay.io'
    resource_definitions = {
        'query': QueryResource,
        'about': AboutResource
    }
    query: QueryResource
    about: AboutResource

    def __init__(self):
        """Create an AnlayticsService."""
        super().__init__(
            json_encode_body=True,
            params={
                'api_version': '0.19',
                # 'render.mode': 'RENDER_MODE_SERIES', // conflicts with 'Accept: text/csv' !! bug?
                # 'json_validate_request': True
            }
        )



"""Interactive authentication callback for client credentials."""

from typing import Optional
import logging
import re

import urllib.parse

from getpass import getpass
import httpx

from .auth import (
    WaylayCredentials,
    ClientCredentials,
    DEFAULT_ACCOUNTS_URL,
    AuthError
)

_http = httpx

log = logging.getLogger(__name__)


def tell(message: str):
    """Show an interactive authentication message."""
    print(message)


def ask(prompt: str, secret: bool = False) -> str:
    """Prompt user for credential information."""
    if secret:
        return getpass(prompt=prompt)
    return input(prompt)


def request_client_credentials_interactive(default_accounts_url: str = DEFAULT_ACCOUNTS_URL) -> WaylayCredentials:
    """Asks interactively for client credentials.

    Default callback provider for an interactive WaylayConfig.
    """
    tell("Authenticating to the Waylay Platform")

    accounts_url: str = default_accounts_url
    acc_validated = False
    while not acc_validated:
        tell(f'Using authentication provider at [{accounts_url}]')
        accounts_url = ask(
            '> alternative endpoint (press enter to continue)?: '
        ) or accounts_url
        accounts_url = _root_url_for(accounts_url)
        accounts_status_resp = httpx.get(accounts_url)
        acc_validated = not accounts_status_resp.is_error
        if acc_validated:
            tell(f"Authenticating at '{accounts_status_resp.json()['name']}'")
        else:
            tell(f"Cannot connect to '{accounts_url}': {accounts_status_resp.reason_phrase}")

    tell("Please provide client credentials for the waylay data client.")
    credentials = ClientCredentials(api_key='', api_secret='', accounts_url=accounts_url)
    retry = 0
    while not credentials.is_well_formed() and retry < 3:
        api_key = ask(prompt='> apiKey : ', secret=False).strip()
        api_secret = ask(prompt='> apiSecret : ', secret=True).strip()
        credentials = ClientCredentials(
            api_key=api_key, api_secret=api_secret, accounts_url=accounts_url
        )
        if not credentials.is_well_formed():
            retry += 1
            if retry >= 3:
                tell('Too many attempts, failing authentication')
                raise AuthError('Too many attempts, failing authentication')
            tell('Invalid apiKey or apiSecret, please retry')
    return credentials


def request_store_config_interactive(profile, save_callback):
    """Save interactively the storage of credentials as a profile."""
    store = ''
    while not store or store[0].lower() not in ('n', 'y', 't', 'f'):
        store = ask(
            f'> Do you want to store these credentials with profile={profile}? [Y]: '
        ) or 'Y'
    if store[0].lower() in ('y', 't'):
        save_location = save_callback()
        tell(
            f"Credential configuration stored as \n\t{save_location}\n"
            "Please make sure this file is treated securely.\n"
            "If compromised, _Revoke_ the api-key on the Waylay Console!"
        )


def _root_url_for(host_or_url: str) -> str:
    scheme, loc, path, query, fragment = urllib.parse.urlsplit(host_or_url)

    if not scheme and not loc:
        # make sure any host name is converted in a https:// url
        scheme = 'https'
        loc = path
        path = ''

    if path.endswith('/'):
        # tenant settings root urls are without trailing slash
        log.warning('Trailing slashes trimmed: %s', host_or_url)
        path = path.rstrip('/')

    return urllib.parse.urlunsplit([scheme, loc, path, query, fragment])



"""REST definitions for status of Waylay BYOML service."""

from .._base import WaylayResource
from .._decorators import (
    exception_decorator,
    return_body_decorator
)

DEFAULT_DECORATORS = [exception_decorator, return_body_decorator]


class AboutResource(WaylayResource):
    """Static status endpoint."""

    actions = {
        'health': {
            'method': 'GET',
            'url': '/',
            'decorators': DEFAULT_DECORATORS,
            'description': 'Get the health status of the <em>BYOML Service</em>'
        }
    }




"""REST client for the Waylay Time Series Analytics Service."""

from .._base import WaylayService

from .model import ModelResource
from .about import AboutResource


class ByomlService(WaylayService):
    """REST client for the Waylay BYOML Service."""

    service_key = 'byoml'
    config_key = 'byoml'
    default_root_url = 'https://byoml.waylay.io'

    resource_definitions = {
        'model': ModelResource,
        'about': AboutResource
    }

    model: ModelResource
    about: AboutResource



"""REST client for the Waylay Platform."""

from collections import defaultdict
from typing import (
    Optional, TypeVar, List, Mapping, Type, Iterable, Dict
)

import pkg_resources

from .service import (
    WaylayService,
    WaylayServiceContext,
    AnalyticsService,
    ByomlService,
    TimeSeriesService,
    ApiService,
    StorageService,
    UtilService,
    ETLService,
    DataService,
    SERVICES
)
from .exceptions import ConfigError

from .config import (
    WaylayConfig,
    DEFAULT_PROFILE
)
from .auth import (
    WaylayCredentials,
    ClientCredentials,
    TokenCredentials,
    DEFAULT_ACCOUNTS_URL
)

import logging

S = TypeVar('S', bound=WaylayService)
logger = logging.getLogger(__name__)


class WaylayClient():
    """REST client for the Waylay Platform."""

    analytics: AnalyticsService
    byoml: ByomlService
    config: WaylayConfig
    timeseries: TimeSeriesService
    api: ApiService
    storage: StorageService
    util: UtilService
    etl: ETLService
    data: DataService

    @classmethod
    def from_profile(
        cls, profile: str = DEFAULT_PROFILE,
        interactive=True, accounts_url=DEFAULT_ACCOUNTS_URL
    ):
        """Create a WaylayClient named profile.

        Uses credentials from environment variables or a locally stored configuration.
        """
        return cls(WaylayConfig.load(profile, interactive=interactive, accounts_url=accounts_url))

    @classmethod
    def from_client_credentials(
        cls, api_key: str, api_secret: str, accounts_url: str = DEFAULT_ACCOUNTS_URL
    ):
        """Create a WaylayClient using the given client credentials."""
        return cls.from_credentials(ClientCredentials(api_key, api_secret, accounts_url))

    @classmethod
    def from_token(
        cls, token_string: str, accounts_url: str = DEFAULT_ACCOUNTS_URL
    ):
        """Create a WaylayClient using a waylay token."""
        return cls.from_credentials(TokenCredentials(token_string, accounts_url))

    @classmethod
    def from_credentials(
        cls, credentials: WaylayCredentials
    ):
        """Create a WaylayClient using the given client credentials."""
        return cls(WaylayConfig(credentials))

    def __init__(self, config: WaylayConfig):
        """Create a WaylayConfig instance."""
        self._services: List[WaylayService] = []
        self.config = config
        self.load_services()

    @property
    def services(self) -> List[WaylayService]:
        """Get the services that are available through this client."""
        return self._services

    @property
    def service_context(self) -> WaylayServiceContext:
        """Get the WaylayServiceContext view on this client."""
        return self

    def configure(self, config: WaylayConfig):
        """Update this client with the given configuration."""
        self.config = config
        for srv in self._services:
            srv.configure(self.config, context=self.service_context)

    def list_root_urls(self) -> Mapping[str, Optional[str]]:
        """List the currently configured root url for each of the registered services."""
        return {srv.config_key: srv.get_root_url() for srv in self._services}

    def __repr__(self):
        """Get a technical string representation of this instance."""
        return (
            f"<{self.__class__.__name__}("
            f"services=[{','.join(list(srv_class.service_key for srv_class in self.services))}],"
            f"config={self.config}"
            ")>"
        )

    def load_services(self):
        """Load all services that are registered with the `waylay_services` entry point."""
        registered_service_classes = [
            srv_class
            for entry_point in pkg_resources.iter_entry_points('waylay_services')
            for srv_class in entry_point.load()
        ]
        if not registered_service_classes:
            logger.warning(
                f"The package {__package__} "
                "seems not to be installed properly. "
                "If it was installed during this runtime session, "
                "please restart the runtime before continuing."
            )
            registered_service_classes = SERVICES
        self.register_service(*registered_service_classes)

    def register_service(self, *service_class: Type[S]) -> Iterable[S]:
        """Create and initialize one or more service of the given class.

        Replaces any existing with the same service_key.
        """
        new_services = [srv_class() for srv_class in service_class]
        new_plugin_priorities: Dict[str, int] = defaultdict(int)
        for srv in new_services:
            new_plugin_priorities[srv.service_key] = max(
                srv.plugin_priority, new_plugin_priorities[srv.service_key]
            )

        # delete existing services
        to_delete_service_index = [
            idx for idx, srv in enumerate(self._services)
            if (
                srv.service_key in new_plugin_priorities and
                srv.plugin_priority <= new_plugin_priorities[srv.service_key]
            )
        ]

        for idx in reversed(to_delete_service_index):
            # delete indexed entries in list from the back
            del self._services[idx]

        # change service list
        for srv in new_services:
            if srv.plugin_priority == new_plugin_priorities[srv.service_key]:
                self._services.append(srv)
                setattr(self, srv.service_key, srv)

        # reconfigure
        self.configure(self.config)
        return new_services

    # implements WaylayServiceContext protocol
    def get(self, service_class: Type[S]) -> Optional[S]:
        """Get the service instance for the provided class, if it is registered.

        Implements the `WaylayServiceContext.get` protocol.
        """
        for srv in self._services:
            if isinstance(srv, service_class):
                return srv
        return None

    def require(self, service_class: Type[S]) -> S:
        """Get the service instance for the given class or raise a ConfigError.

        Implements the `WaylayServiceContext.require` protocol.
        """
        srv = self.get(service_class)
        if srv is None:
            raise ConfigError(f"service {service_class.__name__} is not available.")
        return srv

    def list(self) -> List[WaylayService]:
        """List all registered Services.

        Implements the `WaylayServiceContext.list` protocol.
        """
        return list(self._services)



"""REST definitions for the 'model' entity of the 'byoml' service."""

from zipfile import ZipFile
from io import BytesIO
from functools import wraps
from typing import Callable, List, Any, Union, Tuple, Dict, Optional
import os
import tempfile

import joblib
import pandas as pd
import numpy as np

from .._base import WaylayResource
from .._decorators import (
    return_body_decorator,
    return_path_decorator,
    suppress_header_decorator
)
from ._decorators import (
    byoml_exception_decorator,
    byoml_retry_decorator
)
from ...exceptions import (
    RestRequestError,
)
from ._exceptions import (
    ByomlValidationError,
)


def _input_data_as_list(input_data):
    if isinstance(input_data, list):
        if not input_data:
            # empty list
            return input_data

        if hasattr(input_data[0], 'tolist'):
            # list of numpy arrays
            return [d.tolist() for d in input_data]

        if isinstance(input_data[0], (pd.DataFrame, pd.Series)):
            # list of pandas
            return [d.values.tolist() for d in input_data]

        # list of (list of ...) value types?
        return input_data

    # pandas
    if isinstance(input_data, (pd.DataFrame, pd.Series)):
        return input_data.values.tolist()

    # numpy arrays
    if hasattr(input_data, 'tolist'):
        return input_data.tolist()

    raise RestRequestError(
        f'input data of unsupported type {type(input_data)}'
    )


def model_execution_request_decorator(action_method):
    """Decorate an action to prepare the execution of the model.

    Transforms any input data into a list, and provides it
    as `instances` in the request body.
    """
    @wraps(action_method)
    def wrapped(model_name, input_data, **kwargs):
        body = kwargs.pop('body', {})
        if 'instances' not in body:
            body = {
                'instances': _input_data_as_list(input_data),
                **body
            }
        return action_method(
            model_name,
            body=body,
            **kwargs
        )
    return wrapped


DEFAULT_DECORATORS = [
    byoml_exception_decorator,
    return_body_decorator,
]


def _get_model_decorator() -> List[Callable]:
    return [
        byoml_retry_decorator,
        byoml_exception_decorator,
        return_body_decorator,
    ]


def _execute_model_decorators(response_key: str) -> List[Callable]:
    return [
        byoml_retry_decorator,
        byoml_exception_decorator,
        model_execution_request_decorator,
        return_path_decorator(
            [response_key],
            default_response_constructor=np.array
        )
    ]


DEFAULT_BYOML_MODEL_TIMEOUT = 60

# type aliases for documentation purpose
PathLike = Union[str, os.PathLike]
PytorchModel = Any
TensorflowModel = Any
XgboostModel = Any
SklearnModel = Any
ByomlModel = Union[PytorchModel, TensorflowModel, XgboostModel, SklearnModel]
ModelSerializer = Callable[[PathLike, ByomlModel], PathLike]


def _serialize_torch(work_dir: PathLike, trained_model: PytorchModel) -> PathLike:
    """Serialize a pytorch model to a `model.pt` file."""
    # assuming a TorchScript model
    import torch  # pylint: disable=import-error

    model_file = f'{work_dir}/model.pt'
    model_script = torch.jit.script(trained_model)
    model_script.save(model_file)
    return model_file


def _serialize_joblib(work_dir: PathLike, trained_model: SklearnModel) -> PathLike:
    """Serialize a sklearn model to a `model.joblib` file."""
    model_file = f'{work_dir}/model.joblib'
    joblib.dump(trained_model, model_file)
    return model_file


def _serialize_tf(work_dir: PathLike, trained_model: TensorflowModel) -> PathLike:
    """Serialize a tensorflow model to a model folder."""
    import tensorflow as tf  # pylint: disable=import-error

    tf.saved_model.save(trained_model, work_dir)
    return work_dir


def _serialize_bst(work_dir: PathLike, trained_model: XgboostModel) -> PathLike:
    """Serialize a xgboost model to a `model.bst` file."""
    if hasattr(trained_model, 'save_model'):
        model_file = f'{work_dir}/model.bst'
        trained_model.save_model(model_file)
        return model_file
    raise ByomlValidationError('Could not serialise this model: missing `save_model` method.')


SUPPORTED_FRAMEWORKS: Dict[str, ModelSerializer] = {
    "pytorch": _serialize_torch,
    "sklearn": _serialize_joblib,
    "tensorflow": _serialize_tf,
    "xgboost": _serialize_bst,
}


def assert_dir_exists(dir_name: PathLike):
    """Raise error if the input directory does not exist."""
    if not os.path.exists(str(dir_name)):
        raise ByomlValidationError(f"The directory '{dir_name}' does not exist.")


class ModelResource(WaylayResource):
    """REST Resource for the 'model' entity of the 'byoml' service."""

    link_roots = {
        'doc': '${doc_url}/api/byoml/',
        'iodoc': '${iodoc_url}/api/byoml/?id='
    }

    actions = {
        'list': {
            'method': 'GET',
            'url': '/models',
            'decorators': [
                byoml_exception_decorator,
                return_path_decorator(['available_models'])
            ],
            'description': 'List the metadata of the deployed <em>BYOML Models</em>',
            'links': {
                'doc': '#overview-of-the-api',
                'iodoc': 'overview-of-the-api'
            },
        },
        'list_names': {
            'method': 'GET',
            'url': '/models',
            'decorators': [
                byoml_exception_decorator,
                return_path_decorator(['available_models', 'name'])
            ],
            'description': 'List the names of deployed <em>BYOML Models</em>',
            'links': {
                'doc': '#overview-of-the-api',
                'iodoc': 'overview-of-the-api'
            },
        },
        '_create': {
            'method': 'POST',
            'url': '/models',
            'decorators': DEFAULT_DECORATORS,
            'description': (
                'Build and create a new <em>BYOML Model</em> as specified in the request'
            ),
            'name': 'upload',
            'links': {
                'doc': '#how-to-upload-your-model',
                'iodoc': 'how-to-upload-your-model'
            },
        },
        '_replace': {
            'method': 'PUT',
            'url': '/models/{}',
            'name': 'replace',
            'decorators': DEFAULT_DECORATORS,
            'description': 'Build and replace the named <em>BYOML Model</em>',
            'links': {
                'doc': '#overwriting-a-model',
                'iodoc': 'overwriting-a-model'
            },
        },
        'get': {
            'method': 'GET',
            'url': '/models/{}',
            'decorators': _get_model_decorator(),
            'description': 'Fetch the metadata of the named <em>BYOML Model</em>',
            'links': {
                'doc': '#checking-out-your-model',
                'iodoc': 'checking-out-your-model'
            },
        },
        'get_content': {
            'method': 'GET',
            'url': '/models/{}/content',
            'decorators': [byoml_exception_decorator],
            'description': 'Fetch the content of the named <em>BYOML Model</em>',
            'links': {
                'doc': '#checking-out-your-model',
                'iodoc': 'checking-out-your-model'
            },
        },
        'examples': {
            'method': 'GET',
            'url': '/models/{}/examples',
            'decorators': [
                byoml_exception_decorator,
                return_path_decorator(['example_payloads'])
            ],
            'description': (
                'Fetch the <em>example request input</em> of the named <em>BYOML Model</em>'
            ),
            'links': {
                'doc': '#example-input',
                'iodoc': 'example-input'
            },
        },
        'predict': {
            'method': 'POST',
            'url': '/models/{}/predict',
            'decorators': _execute_model_decorators('predictions'),
            'description': (
                'Execute the <em>predict</em> capability of the named <em>BYOML Model</em>'
            ),
            'links': {
                'doc': '#predictions',
                'iodoc': 'predictions'
            },
        },
        'regress': {
            'method': 'POST',
            'url': '/models/{}/regress',
            'decorators': _execute_model_decorators('result'),
            'description': (
                'Execute the <em>regress</em> capability of the named  <em>BYOML Model</em>'
            ),
            'links': {
                'doc': '#predictions',
                'iodoc': 'predictions'
            },
        },
        'classify': {
            'method': 'POST',
            'url': '/models/{}/classify',
            'decorators': _execute_model_decorators('result'),
            'description': (
                'Execute the <em>classification</em> capability of the named <em>BYOML Model</em>'
            ),
            'links': {
                'doc': '#predictions',
                'iodoc': 'predictions'
            },
        },
        'remove': {
            'method': 'DELETE',
            'url': '/models/{}',
            'decorators': DEFAULT_DECORATORS,
            'description': 'Remove the named <em>BYOML Model</em>',
            'links': {
                'doc': '#deleting-a-model',
                'iodoc': 'deleting-a-model'
            },
        },
    }

    def __init__(self, *args, **kwargs):
        """Create a ModelResource."""
        kwargs.pop('timeout', None)
        super().__init__(*args, timeout=DEFAULT_BYOML_MODEL_TIMEOUT, **kwargs)

    def _get_files_to_zip(self, file_or_dir: PathLike) -> List[Tuple[str, str]]:
        """Get the filenames to zip and the name it should have in the zipfile."""
        file_or_dir = str(file_or_dir)
        if not os.path.isdir(file_or_dir):
            # single file
            zip_file_name = os.path.basename(file_or_dir)
            return [(file_or_dir, zip_file_name)]

        file_names: List[Tuple[str, str]] = []
        for root, _, files in os.walk(file_or_dir):
            for file_name in files:
                # the root will always contain the same suffix, which should not end up in the zip
                zip_root = root[len(file_or_dir):]
                zip_file_name = os.path.join(zip_root, file_name)

                file_path = os.path.join(root, file_name)

                file_names.append((file_path, zip_file_name))

        return file_names

    def _send_model_arguments(
        self, model_name: str, trained_model: Union[PathLike, ByomlModel],
        framework: str = "sklearn", description: str = "",
        work_dir: Optional[PathLike] = None
    ):
        """Upload a binary model with given name, framework and description."""
        if work_dir:
            model_zip_buffer = self._save_model_in_dir(trained_model, work_dir, framework)
        else:
            with tempfile.TemporaryDirectory() as work_dir:
                model_zip_buffer = self._save_model_in_dir(trained_model, work_dir, framework)
        return {
            'body': {
                "name": model_name,
                "framework": framework,
                "description": description
            },
            'files': {
                "file": ('model.zip', model_zip_buffer.getvalue())
            },
        }

    def _save_model_in_dir(
        self, trained_model: Union[PathLike, ByomlModel], work_dir: PathLike, framework: str
    ):
        if not isinstance(trained_model, (str, os.PathLike)):
            file_name = self._serialize_model(trained_model, work_dir, framework)
            files = self._get_files_to_zip(file_name)
        else:
            files = self._get_files_to_zip(trained_model)

        model_zip_buffer = BytesIO()
        with ZipFile(model_zip_buffer, 'w') as zipper:
            for file_name, zip_file_name in files:
                zipper.write(file_name, zip_file_name)

        return model_zip_buffer

    def _serialize_model(
        self, trained_model: ByomlModel, work_dir: PathLike, framework: str
    ) -> PathLike:
        assert_dir_exists(work_dir)

        framework_function = SUPPORTED_FRAMEWORKS.get(framework, None)
        if framework_function is not None:
            return framework_function(work_dir, trained_model)

        raise ByomlValidationError(
            f'Passing a model instance is not supported for this `{framework}` model, '
            'please provide the path to the saved model instead.'
        )

    @suppress_header_decorator('Content-Type')
    def upload(
        self,
        model_name: str,
        trained_model: Union[PathLike, ByomlModel],
        framework: str = "sklearn",
        description: str = "",
        work_dir: Optional[PathLike] = None,
        **kwargs
    ) -> Any:
        """Upload a new machine learning model with given name, framework and description.

        Parameters:
            model_name      The name of the model.
            trained_model   The model object (will be serialised to an zip archive before upload),
                            or a file path to the serialized model file or folder.
            framework       One of the supported frameworks (default 'sklearn').
            description     Description of the model.
            work_dir        Optional location of the working directory used to serialize the model.
                            If not specified, a temporary directory is used.
            (other args)    Passed onto the underlying 'POST /model/{name}' request
        """
        return self._create(  # pylint: disable=no-member
            **self._send_model_arguments(
                model_name, trained_model,
                framework=framework,
                description=description,
                work_dir=work_dir,
            ),
            **kwargs
        )

    @suppress_header_decorator('Content-Type')
    def replace(
        self,
        model_name: str,
        trained_model: Union[PathLike, ByomlModel],
        framework: str = "sklearn",
        description: str = "",
        work_dir: Optional[PathLike] = None,
        **kwargs
    ) -> Any:
        """Replace a machine learning model with given name, framework and description.

        Parameters:
            model_name      The name of the model.
            trained_model   The model object (will be serialised to an zip archive before upload).
            framework       One of the supported frameworks (default 'sklearn').
            description     Description of the model.
            work_dir        Optional location of the working directory used to serialize the model.
                            If not specified, a temporary directory is used.
            (other)         Passed onto the underlying 'PUT /model/{name}' request.
        """
        return self._replace(   # pylint: disable=no-member
            model_name,
            **self._send_model_arguments(
                model_name, trained_model,
                framework=framework,
                description=description,
                work_dir=work_dir,
            ),
            **kwargs
        )



"""Base classes for Waylay REST Services and Resources."""
__docformat__ = "google"

from typing import (
    Optional, Type, TypeVar,
    Mapping, List, Dict, Optional,
    Callable
)
try:
    from typing import Protocol
except ImportError:
    # typing.Protocol is a 3.8 feature ...
    # ... but typing_extensions provides forward compatibility.
    from typing_extensions import Protocol  # type: ignore

from string import Template

from simple_rest_client.api import Resource, API
from waylay.config import WaylayConfig
from waylay.exceptions import ConfigError

S = TypeVar('S', bound='WaylayService')
R = TypeVar('R', bound='WaylayResource')


class WaylayServiceContext(Protocol):
    """View protocol for the dynamic service context."""

    def get(self, service_class: Type[S]) -> Optional[S]:
        """Get the service instance for the given class, if available."""

    def require(self, service_class: Type[S]) -> S:
        """Get the service instances for the given class or raise a ConfigError."""

    def list(self) -> List['WaylayService']:
        """List all available service instances."""


DEFAULT_SEVICE_TIMEOUT = 10


# Implementation Note
# -------------------
# Current solution is scaffolded around the `simple_rest_client`, but is planned to be refactored.

class WaylayAction(dict):
    """Client object representing a Waylay REST action."""

    @property
    def resource(self) -> 'WaylayResource':
        """Get the parent resource of this action."""
        return self['_resource']

    @property
    def id(self) -> str:
        """Get the action id."""
        return self['id']

    @property
    def name(self) -> str:
        """Get the action name as exposed in the SDK."""
        return self.get('name', self.id)

    @property
    def method(self) -> Optional[str]:
        """Get the action HTTP method."""
        return self.get('method')

    @property
    def url(self) -> Optional[str]:
        """Get the action url template."""
        return self.get('url')

    @property
    def description(self) -> Optional[str]:
        """Get the action description."""
        return self.get('description')

    @property
    def hal_links(self) -> Dict[str, Dict[str, str]]:
        """Get (documentation) links templates."""
        roots = self.resource.hal_roots
        return {
            link: dict(href=roots.get(link, '') + ref)
            for link, ref in self.get('links', {}).items()
        }

    @property
    def action_method(self) -> Callable:
        """Get the SDK action method that is used to expose this action."""
        return getattr(self.resource, self.id)

    @property
    def sdk_action_method(self) -> Callable:
        """Get the REST action method that is used to expose this action."""
        return getattr(self.resource, self.name)

    @property
    def sdk_action_doc(self) -> str:
        """Get the sdk python documentation for this action."""
        doc = self.sdk_action_method.__doc__ or ''
        if self.name != self.id:
            # wrapped actions
            doc += "\n" + (self.action_method.__doc__ or '')
        return doc


class WaylayResource(Resource):
    """Client object representing a Waylay REST Resource.

    This is a collection of REST operations that have a single Waylay Entity as subject.
    """

    service: 'WaylayService'

    # class variables
    link_roots: Dict[str, str] = {}

    def __init__(self, *args, **kwargs):
        """Create a Waylay Resource."""
        self.actions = {
            id: WaylayAction(id=id, _resource=self, **action)
            for id, action in self.actions.items()
        }
        super().__init__(*args, **kwargs)

    @property
    def name(self):
        """Get the name that identifies this resource in the Python SDK."""
        return self.resource_name

    @property
    def description(self):
        """Get a description of this service."""
        return self.__doc__

    def add_action(self, action_name: str):
        """Add action, and apply decorators."""
        super().add_action(action_name)
        self.decorate_action(action_name)

    def decorate_action(self, action_name):
        """Decorate the action if a 'decorators' definition exist."""
        action = self.get_action(action_name)
        decorators = action.get('decorators', None)
        if decorators:
            for decorator in decorators:
                setattr(self, action_name, decorator(getattr(self, action_name)))

    def __repr__(self):
        """Get a technical string representation of this action."""
        actions_repr = ', '.join(
            f"{name}: '{action_def['method']} {action_def['url']}'"
            for name, action_def in self.actions.items()
        )
        return (
            f"<{self.__class__.__name__}("
            f"actions=[{actions_repr}]"
            ")>"
        )

    @property
    def hal_roots(self) -> Dict[str, str]:
        """Get the root urls for documentation links if this resource."""
        return {
            link: Template(
                href_root
            ).safe_substitute(
                root_url=self.api_root_url,
                doc_url=self.service.config.DOC_URL,
                iodoc_url=self.service.config.IODOC_URL,
            )
            for link, href_root in self.link_roots.items()
        }

    def hal_links(self, action: Optional[str]) -> Dict[str, Dict[str, str]]:
        """Create a HAL `_links` representation for (documentation) links.

        Attributes:
            action      if specified, give links for a specific action rather than the resource.
        """
        hrefs = {rel: '' for rel in self.link_roots}
        if action:
            hrefs = self.actions.get(action).links

        return {
            link_rel: dict(
                href=Template(
                    href_root + hrefs.get(link_rel, '')
                ).safe_substitute(
                    root_url=self.api_root_url,
                    doc_url=self.service.config.DOC_URL,
                    iodoc_url=self.service.config.IODOC_URL,
                )
            )
            for link_rel, href_root in self.link_roots.items()
            if link_rel in hrefs
        }


class WaylayService(API):
    """Client object representing a Waylay Service.

    A collection of Resources with their operations.
    """

    config: WaylayConfig

    # class variables
    service_key: str = ''
    config_key: str = 'api'
    default_root_url: Optional[str] = None
    resource_definitions: Mapping[str, Type[Resource]] = {}
    plugin_priority = 0
    link_templates: Dict[str, str] = {}

    @property
    def name(self):
        """Get the name that identifies this service in the Python SDK."""
        return self.service_key

    @property
    def description(self):
        """Get a description of this service."""
        return self.__doc__

    @property
    def resources(self):
        """Get the REST resources supported by this service."""
        return self.list_resources()

    def __init__(self, *args, **kwargs):
        """Create a WaylayService."""
        timeout = kwargs.pop('timeout', DEFAULT_SEVICE_TIMEOUT)
        json_encode_body = kwargs.pop('json_encode_body', True)
        super().__init__(*args, timeout=timeout, json_encode_body=json_encode_body, **kwargs)
        for name, resource_class in self.resource_definitions.items():
            self._add_waylay_resource(resource_name=name, resource_class=resource_class)

    def _add_waylay_resource(self, resource_name: str, resource_class: Type[R], **kwargs) -> R:
        self.add_resource(resource_name=resource_name, resource_class=resource_class, **kwargs)
        waylay_resource: R = self._resources[self.correct_attribute_name(resource_name)]
        waylay_resource.service = self
        return waylay_resource

    def set_root_url(self, root_url):
        """Set the root url and reconfigure the service."""
        self.config.set_root_url(self.config_key, root_url)
        self.reconfigure()

    def get_root_url(self) -> Optional[str]:
        """Get the root url."""
        if self.config is None:
            return self.default_root_url
        return self.config.get_root_url(self.config_key, self.default_root_url)

    @property
    def root_url(self) -> Optional[str]:
        """Get the root url."""
        return self.get_root_url()

    def configure(self: S, config: WaylayConfig, context: WaylayServiceContext) -> S:
        """Configure endpoints and authentication with given configuration.

        Returns self
        """
        self.config = config
        return self.reconfigure()

    def reconfigure(self: S) -> S:
        """Configure endpoints and authentication with current configuration.

        Returns self
        """
        if self.config is None:
            return self
        root_url = self.get_root_url()
        if not root_url:
            raise ConfigError(
                f'The service `{self.service_key}` has no url configuration. '
                f'Please provide a endpoint using a setting with key `{self.config_key}`,'
                'or request your tenant administrator '
                f'to configure the global setting `waylay_{self.config_key}`.'
            )
        for resource in self._resources.values():
            resource.api_root_url = root_url
            resource.client.auth = self.config.auth
        return self

    def list_resources(self) -> List[WaylayResource]:
        """List the WaylayResources of this service."""
        return list(self._resources.values())

    def __repr__(self):
        """Get a technical string representation of this service."""
        return (
            f"<{self.__class__.__name__}("
            f"service_key={self.service_key},"
            f"config_key={self.config_key},"
            f"root_url={self.get_root_url()},"
            f"resources=[{', '.join(self._resources.keys())}]"
            ")>"
        )

    def hal_links(self) -> Dict[str, Dict[str, str]]:
        """Create a HAL `_links` representation for the documentation links."""
        return {
            rel: dict(href=Template(href).safe_substitute(
                root_url=self.root_url,
                doc_url=self.config.DOC_URL,
                iodoc_url=self.config.IODOC_URL,
            ))
            for rel, href in self.link_templates.items()
        }



"""Exceptions specific to the Analytics Service."""

from typing import (
    List, Mapping, Any
)
from ...exceptions import RestResponseError, RestResponseParseError


class AnalyticsActionError(RestResponseError):
    """Error that represents the json messages of a analytics response."""

    @property
    def messages(self) -> List[Mapping[str, Any]]:
        """Get the list of message objects returned by an analytics error response."""
        return self._get_from_body('messages', [])

    @property
    def message(self):
        """Get the main user error returned by an analytics error response."""
        return self._get_from_body('error', super().message)


class AnalyticsActionParseError(RestResponseParseError, AnalyticsActionError):
    """Indicates that a analytics response could not be parsed."""



"""REST definitions for the Settings entity of the Waylay Engine Service."""

from waylay.service import WaylayResource
from waylay.service import decorators


DEFAULT_DECORATORS = [decorators.exception_decorator, decorators.return_body_decorator]


class SettingsResource(WaylayResource):
    """REST Resource for the `settings` entity of the provisioning API."""

    actions = {
        'get': {
            'method': 'GET', 'url': '/api/settings',
            'decorators': DEFAULT_DECORATORS,
            'description': 'Retrieve tenant global settings.',
            'links': {
            },
        },
    }



"""Waylay Rest Services."""

from ._base import WaylayService, WaylayResource, WaylayServiceContext
from . import _decorators as decorators

from .analytics import AnalyticsService
from .byoml import ByomlService
from .timeseries import TimeSeriesService
from .api import ApiService
from .storage import StorageService
from .util import UtilService
from .etl import ETLService
from .data import DataService


SERVICES = [
    AnalyticsService,
    ByomlService,
    TimeSeriesService,
    ApiService,
    StorageService,
    UtilService,
    ETLService,
    DataService
]



"""Definitions for the 'query' entity of the 'analytics' Service."""

from .._base import WaylayResource
from .._decorators import (
    return_path_decorator
)
from ._decorators import (
    analytics_exception_decorator,
    analytics_return_dataframe_decorator,
    MultiFrameHandling
)


CONFIG_ENTITY_DECORATORS = [
    analytics_exception_decorator,
    return_path_decorator(['query'])
]

CONFIG_LIST_DECORATORS = [
    analytics_exception_decorator,
    return_path_decorator(['queries', 'name'])
]

DATA_RESPONSE_DECORATORS = [
    analytics_exception_decorator,
    analytics_return_dataframe_decorator(
        'data',
        default_frames_handling=MultiFrameHandling.JOIN
    )
]

CONFIG_STATUS_DECORATORS = [
    analytics_exception_decorator,
    return_path_decorator([])
]


class QueryResource(WaylayResource):
    """REST Resource for the 'query' entity of the 'analytics' Service."""

    link_roots = {
        'doc': '${doc_url}/api/query/',
        'iodoc': '${iodoc_url}/api/query/',
        'apidoc':  '${root_url}/apidocs/'
    }

    actions = {
        'list': {
            'method': 'GET',
            'url': '/config/query',
            'decorators': CONFIG_LIST_DECORATORS,
            'description': 'List or search <em>Query Configurations</em>.',
            'links': {
                'doc': 'search/#search',
                'iodoc': '?id=data-query-search-api',
                'apidoc': '#/query%20config/get_config_query'
            }
        },
        'create': {
            'method': 'POST',
            'url': '/config/query',
            'decorators': CONFIG_ENTITY_DECORATORS,
            'description': 'Create a new <em>Query Configuration</em>.',
            'links': {
                'doc': 'crud/#create',
                'iodoc': '?id=create',
                'apidoc': '#/query%20config/post_config_query'
            }
        },
        'get': {
            'method': 'GET',
            'url': '/config/query/{}',
            'decorators': CONFIG_ENTITY_DECORATORS,
            'description': 'Fetch the named <em>Query Configuration</em>.',
            'links': {
                'doc': 'crud/#retrieve',
                'iodoc': '?id=retrieve',
                'apidoc': '#/query%20config/get_config_query__query_name_'
            }
        },
        'remove': {
            'method': 'DELETE',
            'url': '/config/query/{}',
            'decorators': CONFIG_STATUS_DECORATORS,
            'description': 'Remove the named <em>Query Configuration</em>.',
            'links': {
                'doc': 'crud/#delete',
                'iodoc': '?id=delete',
                'apidoc': '#/query%20config/delete_config_query__query_name_'
            }
        },
        'replace': {
            'method': 'PUT',
            'url': '/config/query/{}',
            'decorators': CONFIG_ENTITY_DECORATORS,
            'description': 'Replace the named <em>Query Configuration</em>.',
            'links': {
                'doc': 'crud/#replace',
                'iodoc': '?id=replace',
                'apidoc': '#/query%20config/put_config_query__query_name_'
            }
        },
        'data': {
            'method': 'GET',
            'url': '/data/query/{}',
            'decorators': DATA_RESPONSE_DECORATORS,
            'description': (
                'Execute the timeseries query specified by the named <em>Query Configuration</em>.'
            ),
            'links': {
                'doc': 'data/#query-execution',
                'iodoc': '?id=query-execution',
                'apidoc': '#/data/get_data_query__query_name_'
            }
        },
        'execute': {
            'method': 'POST',
            'url': '/data/query',
            'decorators': DATA_RESPONSE_DECORATORS,
            'description': 'Execute the timeseries query specified in the request body.',
            'links': {
                'doc': 'data/#query-execution',
                'iodoc': '?id=query-execution',
                'apidoc': '#/data/post_data_query'
            }
        },
    }



"""Resource action method decorators that are generally usefull."""

from functools import wraps
from typing import (
    Union, List, Mapping, Any, Callable, TypeVar, Dict, Optional
)
from simple_rest_client.exceptions import ErrorWithResponse
from ..exceptions import RestResponseError, RestResponseParseError


def suppress_header_decorator(header_key):
    """Create a decorator that suppresses a configured header on a resource during execution."""
    def decorator(action_method):
        @wraps(action_method)
        def wrapped(slf, *args, **kwargs):
            header_value = slf.headers.pop(header_key, None)
            try:
                return action_method(slf, *args, **kwargs)
            finally:
                if header_value:
                    slf.headers[header_key] = header_value
        return wrapped
    return decorator


def exception_decorator(action_method):
    """Create a decorator that parses json error responses."""
    @wraps(action_method)
    def wrapped(*args, **kwargs):
        try:
            return action_method(*args, **kwargs)
        except ErrorWithResponse as exc:
            raise RestResponseError.from_cause(exc) from exc
    return wrapped


def default_params_decorator(default_params: Dict[str, Any]):
    """Create a decorator that initializes default url request parameters."""
    def decorator(action_method):
        @wraps(action_method)
        def wrapped(slf, *args, **kwargs):
            params = kwargs.pop('params', {})
            for key, value in default_params.items():
                if key not in params:
                    params[key] = value
            kwargs['params'] = params
            return action_method(slf, *args, **kwargs)
        wrapped.__doc__ = (
            (wrapped.__doc__ or '') +
            "Uses default url parameters (`params` argument):" +
            "".join(
                f"\n    '{key}': {default_value}"
                for key, default_value in default_params.items()
            )
        )
        return wrapped
    return decorator


T = TypeVar('T')


def identity_transform(obj: T) -> T:
    """Return the argument."""
    return obj


def return_path_decorator(
    default_path: List[Union[int, str]],
    default_response_constructor: Optional[Callable[[Any], Any]] = None
):
    """Create a decorator method that extracts a part of the json body of an action response.

    The 'default_path' is a list of 'int' or 'str' keys that select through
    'dict' and 'list' structures.
    The option 'response_constructor' wrap the value in a class or constructor method
    It processes the following qualified arguments in a method call:
    * if 'raw=True' is set, the original response is returned
    * if 'select_path' is set, that path is used rather than the 'default_path'
       provided in the decorator constructor.
    * if 'response_constructor' is set, it is used to wrap the response
      rather than 'default_response_constructor'
    """
    def decorator(action_method):
        @wraps(action_method)
        def wrapped(*args, **kwargs):
            raw = kwargs.pop('raw', False)
            select_path = kwargs.pop('select_path', default_path)
            constructor = kwargs.pop(
                'response_constructor', default_response_constructor
            ) or identity_transform
            response = action_method(*args, **kwargs)
            if raw:
                return response
            if response.status_code == 204:
                # No-Content
                return None
            try:
                return constructor(
                    _select_path_from(response.body, [], select_path)
                )
            except AttributeError as exc:
                raise RestResponseParseError(exc.args[0], response) from exc
        resp_constr_arg = 'response_constructor'
        if default_response_constructor:
            resp_constr_arg += f"={_fqn_name(default_response_constructor)}"
        select_path_arg = f"select_path={default_path}"
        raw_arg = 'raw=False'
        dec_doc = f"""
Returns the (parsed json) response body. Arguments that configure response handling:
    {select_path_arg:30}: extract path from the response body.
    {resp_constr_arg:30}: transforms the extracted response.
    {raw_arg:30}: If `True`, returns unmodified http response information.
"""
        wrapped.__doc__ = (wrapped.__doc__ or '') + dec_doc
        return wrapped
    return decorator


def _fqn_name(func: Any) -> str:
    if func.__module__ == 'builtins':
        return func.__qualname__
    return f"{func.__module__}.{func.__qualname__}"


def return_body_decorator(action_method):
    """Create a decorator that returns the body of an action response.

    Skipped if 'raw=True' is provided in the call.
    """
    return return_path_decorator([])(action_method)


def _select_path_from(
    value: Any, ctx_path: List[Union[int, str]], path: List[Union[int, str]]
) -> Any:
    # Recursively select values from 'Mapping' (dict, 'str' key) input,
    # or positional elements from 'List' (list, 'int' key) input,
    # with a 'path' consisting of a list of keys.
    # Applied keys are moved to the 'ctx_path' to be used for error reporting.
    # 'str' keys applied to a 'List' are mapped through to the underlying structure.

    if not path:
        return value
    key = path[0]
    new_ctx_path = ctx_path + [key]
    rem_path = path[1:]
    if isinstance(value, List):
        if isinstance(key, int) and len(value) < key:
            return _select_path_from(value[key], new_ctx_path, rem_path)
        if isinstance(key, str):
            return [
                _select_path_from(element[key], new_ctx_path, rem_path)
                for element in value
            ]
    if isinstance(value, Mapping) and isinstance(key, str) and key in value:
        return _select_path_from(value[key], new_ctx_path, rem_path)

    ctx_path_str = '.'.join(str(p) for p in new_ctx_path)
    raise AttributeError(f"Cannot find '{ctx_path_str}' in response body.")




"""Utility tools and services for the Python SDK."""

from waylay.service import WaylayService, WaylayServiceContext
from waylay.config import WaylayConfig

from .info import InfoTool


class UtilService(WaylayService):
    """Utility Service for the python SDK."""

    service_key = 'util'

    info: InfoTool

    def configure(self, config: WaylayConfig, context: WaylayServiceContext) -> 'UtilService':
        """Configure endpoints and authentication with given config."""
        self.config = config
        self.info = InfoTool(context)
        return self.reconfigure()

    def reconfigure(self) -> 'UtilService':
        """Reconfigure this service after a configuration change."""
        return self



"""REST definitions for the 'subscription' entity of the 'storage' service."""
from urllib.parse import quote

from waylay.service import WaylayResource
from waylay.service import decorators


class SubscriptionResource(WaylayResource):
    """REST Resource for the 'subscription' entity of the 'storage' service."""

    link_roots = {
        'iodoc': '${iodoc_url}/api/storage/?id=',
        'apidoc':  '${root_url}/docs#/subscription/',
        'openapi': '${root_url}/openapi.json'
    }
    actions = {
        'list': {
            'method': 'GET', 'url': '/subscription/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator(['subscriptions']),
            ],
            'description': 'List available subscriptions for a given bucket.',
            'links': {
                'iodoc': 'list-bucket-subscriptions',
                'apidoc': 'query_subscriptions_subscription__bucket_name__get',
                'openapi': f"#/paths/{quote('/subscription/{bucket_name}')}/get"
            }
        },
        'get': {
            'method': 'GET', 'url': '/subscription/{}/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_body_decorator,
            ],
            'description': 'Retrieve the representation of a notification subscription.',
            'links': {
                'iodoc': 'get-subscription',
                'apidoc': 'get_subscription_subscription__bucket_name___id__get',
                'openapi': f"#/paths/{quote('/subscription/{bucket_name}/{id}')}/get"
            }
        },
        'create': {
            'method': 'POST', 'url': '/subscription/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_body_decorator,
            ],
            'description': 'Create a new notification subscription.',
            'links': {
                'iodoc': 'create-subscription',
                'apidoc': 'post_subscription_subscription__bucket_name__post',
                'openapi': f"#/paths/{quote('/subscription/{bucket_name}/{id}')}/post"
            }
        },
        'replace': {
            'method': 'PUT', 'url': '/subscription/{}/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_body_decorator,
            ],
            'description': 'Create or Replace the definition of a notification subscription.',
            'links': {
                'iodoc': 'update-subscription',
                'apidoc': 'put_subscription_subscription__bucket_name___id__put',
                'openapi': f"#/paths/{quote('/subscription/{bucket_name}/{id}')}/put"
            }
        },
        'remove': {
            'method': 'DELETE', 'url': '/subscription/{}/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_body_decorator,
            ],
            'description': 'Remove a notification subscription.',
            'links': {
                'iodoc': 'delete-subscription',
                'apidoc': 'delete_subscription_subscription__bucket_name___id__delete',
                'openapi': f"#/paths/{quote('/subscription/{bucket_name}/{id}')}/delete"
            }

        },
        'remove_all': {
            'method': 'DELETE', 'url': '/subscription/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_body_decorator,
            ],
            'description': 'Remove all notification subscription that satisfy a query.',
            'links': {
                'iodoc': 'delete-subscriptions',
                'apidoc': 'delete_subscriptions_subscription__bucket_name__delete',
                'openapi': f"#/paths/{quote('/subscription/{bucket_name}')}/delete"
            }
        },
    }



"""REST definitions for the 'object' entity of the 'storage' service."""

import isodate
import re
from datetime import timedelta
from urllib.parse import urlsplit, parse_qs, quote
from functools import wraps

from typing import Union, Optional, Dict, Iterator

from waylay.service import WaylayResource
from waylay.service import decorators


def parse_expiry_params(expiry: Union[str, timedelta]) -> Dict[str, int]:
    """Parse sign expiry parameters.

    Input should be either:
    * an ISO period expression
    * a simple string expression `1d` , `6h`, `300s`
    * a `datetime.timedelta` objec
    """
    if expiry is None:
        return {}
    if isinstance(expiry, str):
        if 'P' in expiry:
            return parse_expiry_params(isodate.parse_duration(expiry))
        simple_period_match = re.match(r"(\d+)([dhs])", expiry)
        if simple_period_match:
            amount, dhs_key = simple_period_match.groups()
            key: str = dict(
                d='expiry_days', h='expiry_hours', s='expiry_seconds'
            )[dhs_key]
            return {key: int(amount)}

    if isinstance(expiry, timedelta):
        return {
            'expiry_days': expiry.days,
            'expiry_seconds': expiry.seconds
        }
    raise AttributeError(f'Not an expiry expression: {expiry}')


def _object_path_remove_path_suffix(action_method):
    # remove the `/` suffix from the (virtual folder) object_path,
    # as we already include it in the url template.
    @wraps(action_method)
    def wrapped(*args, **kwargs):
        new_args = args
        if args[1].endswith('/'):
            new_args = [args[1].rstrip('/'), *args[1:]]
        return action_method(*new_args, **kwargs)
    return wrapped


class FolderResource(WaylayResource):
    """REST Resource for the folder 'object' entity of the 'storage' service."""

    link_roots = {
        'iodoc': '${iodoc_url}/api/storage/?id=',
        'apidoc':  '${root_url}/docs#/storage/',
        'openapi': '${root_url}/openapi.json'
    }

    actions = {
        'list': {
            'method': 'GET', 'url': '/bucket/{}/{}/', 'decorators': [
                _object_path_remove_path_suffix,
                decorators.exception_decorator,
                decorators.return_path_decorator(
                    ['objects']
                ),
            ],
            'description': 'List objects in this folder.',
            'links': {
                'iodoc': 'list-object',
                'apidoc': 'list_or_get_objects_bucket__bucket_name___object_path__get',
                'openapi': f"#/paths/{quote('/bucket/{bucket_name}/{object_path}')}/get"
            }
        },
        'create': {
            'method': 'PUT', 'url': '/bucket/{}/{}/', 'decorators': [
                _object_path_remove_path_suffix,
                decorators.exception_decorator,
                decorators.return_body_decorator,
            ],
            'description': 'Create a folder.',
            'links': {
                'iodoc': 'put-folder',
                'apidoc': 'create_folder_bucket__bucket_name___object_path___put',
                'openapi': f"#/paths/{quote('/bucket/{bucket_name}/{object_path}/')}/put"
            }
        },
        'stat': {
            'method': 'GET', 'url': '/bucket/{}/{}/?stat=true', 'decorators': [
                _object_path_remove_path_suffix,
                decorators.exception_decorator,
                decorators.return_body_decorator,
            ],
            'description': 'Get the details of this folder',
            'links': {
                'iodoc': 'get-object',
                'apidoc': 'list_or_get_objects_bucket__bucket_name___object_path__get',
                'openapi': f"#/paths/{quote('/bucket/{bucket_name}/{object_path}')}/get"
            }
        },
        'remove': {
            'method': 'DELETE', 'url': '/bucket/{}/{}/', 'decorators': [
                _object_path_remove_path_suffix,
                decorators.exception_decorator,
                decorators.return_path_decorator(
                    ['_links']
                )
            ],
            'description': 'Remove this folder',
            'links': {
                'iodoc': 'delete-object-or-folder',
                'apidoc': 'delete_object_bucket__bucket_name___object_path__delete',
                'openapi': f"#/paths/{quote('/bucket/{bucket_name}/{object_path}')}/delete"
            }
        },
    }


class ObjectResource(WaylayResource):
    """REST Resource for the 'object' entity of the 'storage' service."""

    link_roots = {
        'iodoc': '${iodoc_url}/api/storage/?id=',
        'apidoc':  '${root_url}/docs#/storage/',
        'openapi': '${root_url}/openapi.json'
    }

    actions = {
        'sign_get': {
            'method': 'GET', 'url': '/bucket/{}/{}?sign=GET', 'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator(
                    ['_links', 'get_object', 'href']
                ),
            ],
            'description': 'Create a signed http GET link for the given bucket and object path.',
            'links': {
                'iodoc': 'sign_url',
                'apidoc': 'list_or_get_objects_bucket__bucket_name___object_path__get',
                'openapi': f"#/paths/{quote('/subscription/{bucket_name}')}/get"
            }
        },
        'sign_post': {
            'method': 'GET', 'url': '/bucket/{}/{}?sign=POST', 'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator(
                    ['_links', 'post_object']
                ),
            ],
            'description': 'Create a signed http POST link for the given bucket and object path.',
            'links': {
                'iodoc': 'sign_url',
                'apidoc': 'list_or_get_objects_bucket__bucket_name___object_path__get',
                'openapi': f"#/paths/{quote('/subscription/{bucket_name}')}/get"
            }
        },
        'sign_put': {
            'method': 'GET', 'url': '/bucket/{}/{}?sign=PUT', 'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator(
                    ['_links', 'put_object', 'href']
                ),
            ],
            'description':  'Create a signed http PUT link for the given bucket and object path.',
            'links': {
                'iodoc': 'sign_url',
                'apidoc': 'list_or_get_objects_bucket__bucket_name___object_path__get',
                'openapi': f"#/paths/{quote('/subscription/{bucket_name}')}/get"
            }
        },
        'stat': {
            'method': 'GET', 'url': '/bucket/{}/{}?stat=true', 'decorators': [
                decorators.exception_decorator,
                decorators.return_body_decorator,
            ],
            'description': 'Get the object metadata for the given bucket and object path.',
            'links': {
                'iodoc': 'get-object',
                'apidoc': 'list_or_get_objects_bucket__bucket_name___object_path__get',
                'openapi': f"#/paths/{quote('/subscription/{bucket_name}')}/get"
            }
        },
        'remove': {
            'method': 'DELETE', 'url': '/bucket/{}/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_body_decorator
            ],
            'description': 'Remove the object at the given bucket and object path.',
            'links': {
                'iodoc': 'delete-object-or-folder',
                'apidoc': 'delete_object_bucket__bucket_name___object_path__delete',
                'openapi': f"#/paths/{quote('/subscription/{bucket_name}')}/get"
            }
        },
        'list': {
            'method': 'GET', 'url': '/bucket/{}/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator(
                    ['objects']
                ),
            ],
            'description': 'List objects in given bucket with a given path prefix.',
            'links': {
                'iodoc': 'list-object',
                'apidoc': 'list_or_get_objects_bucket__bucket_name___object_path__get',
                'openapi': f"#/paths/{quote('/subscription/{bucket_name}')}/get"
            }
        },
    }

    def iter_list_all(self, bucket: str, prefix: str, params: Optional[Dict] = None) -> Iterator[Dict]:
        """Use paging to iterate over all objects."""
        done = False
        while not done:
            result = self.list(bucket, prefix, select_path=[], params=params)  # pylint: disable=no-member
            for object in result.get('objects', []):
                yield object
            next_link = result.get('_links', {}).get('next')
            if next_link:
                query = urlsplit(next_link['href']).query
                params = {k: v[0] for k, v in parse_qs(query, keep_blank_values=True).items()}
            else:
                done = True



"""REST definitions for version status of Waylay Analytics service."""

from .._base import WaylayResource
from .._decorators import (
    exception_decorator,
    return_body_decorator
)

DEFAULT_DECORATORS = [exception_decorator, return_body_decorator]


class AboutResource(WaylayResource):
    """Version information."""

    actions = {
        'version': {
            'method': 'GET',
            'url': '/',
            'decorators': DEFAULT_DECORATORS,
            'description': 'Version info of the <em>Analytics Service</em> at this endpoint.'}
    }



"""Internal service with tools for timeseries import and export."""

from waylay.service import WaylayService, WaylayServiceContext
from waylay.config import WaylayConfig
from waylay.service.storage import StorageService
from waylay.service.etl import ETLService
from waylay.service.api import ApiService
from waylay.service.analytics import AnalyticsService

from .tool import TimeSeriesETLTool


class TimeSeriesService(WaylayService):
    """Tool service the timeseries import and export operations."""

    service_key = 'timeseries'

    etl_tool: TimeSeriesETLTool

    def configure(self, config: WaylayConfig, context: WaylayServiceContext) -> 'TimeSeriesService':
        """Configure endpoints and authentication with given config."""
        self.config = config
        self.etl_tool = TimeSeriesETLTool(
            context.require(StorageService),
            context.require(ETLService),
            context.require(ApiService),
            context.require(AnalyticsService),
        )
        return self.reconfigure()

    def reconfigure(self) -> 'TimeSeriesService':
        """Reconfigure after configuration change."""
        return self



"""Utilities to parse and render ETL timeseries files."""
from typing import (
    Optional, Union, Tuple, Any, Set,
    Sequence,
    List, Callable,
    Iterable,
    Iterator,
    ContextManager,
    cast
)
import gzip
import csv
import sys
import io
import os
from pathlib import Path
from datetime import datetime
from dataclasses import replace
from contextlib import contextmanager

import pandas as pd

from waylay.exceptions import RequestError

from .model import (
    Metric,
    Resource,
    SeriesSettings,
    ETLFile,
    WaylayETLSeriesImport,

    SeriesInput,

    MeasurementValue,
    Measurement,
    MeasurementIterator,

    SeriesIterator,
    ETL_IMPORT_COLUMN_NAMES,

    render_timestamp,
    try_parse_float,
    fromisoformat,
    default_create_timestamp,
    ParserRequestError,
)
from .import_csv import (
    prepare_settings_csv,
    iter_timeseries_csv
)
from .import_dataframe import (
    prepare_settings_dataframe,
    iter_timeseries_dataframe
)
from .etlfile import (
    read_etl_import_as_stream,
    read_etl_import,
    dataframe_from_iterator,
    list_resources,
)


def iter_timeseries(
    input_data: SeriesInput,
    settings: SeriesSettings,
    progress: bool = False
) -> SeriesIterator:
    """Create a SeriesInterator for the given input source and settings."""
    if isinstance(input_data, pd.DataFrame):
        return iter_timeseries_dataframe(input_data, settings, progress)
    else:
        return iter_timeseries_csv(input_data, settings, progress)


def prepare_etl_import(
    series_input: SeriesInput,
    import_file: Optional[ETLFile] = None,
    settings: Optional[SeriesSettings] = None,
    **settings_args: Any
) -> WaylayETLSeriesImport:
    """Update the mapping settings by extracting metadata from an input source."""
    import_file = import_file or ETLFile()
    settings = settings or SeriesSettings()
    if settings_args:
        settings = replace(settings, **settings_args)

    # input settings validation
    etl_import = WaylayETLSeriesImport(
        series_input=series_input,
        settings=settings,
        import_file=import_file
    )

    # input series validation
    if isinstance(series_input, pd.DataFrame):
        settings = prepare_settings_dataframe(series_input, settings)
        etl_import = replace(etl_import, settings=settings)
    else:
        settings = prepare_settings_csv(series_input, settings)
        etl_import = replace(etl_import, settings=settings)
    return etl_import


def create_etl_import(
    etl_import: WaylayETLSeriesImport,
    progress: bool = True
) -> WaylayETLSeriesImport:
    """Create an ETL import file from the given input."""
    file_path = etl_import.import_file.path
    if file_path.exists():
        raise RequestError(
            f'The file {file_path} already exists. '
            'Please remove or specify another etl-import file name.'
        )

    timeseries_data = iter_timeseries(etl_import.series_input, etl_import.settings, progress=progress)
    with gzip.open(file_path, 'wt') as csv_file:
        writer = csv.writer(csv_file, delimiter=',', quoting=csv.QUOTE_MINIMAL)
        writer.writerow(ETL_IMPORT_COLUMN_NAMES)
        for resource, metric, measurements in timeseries_data:
            for timestamp, value in measurements:
                writer.writerow([
                    resource,
                    metric,
                    render_timestamp(timestamp),
                    value
                ])
    return etl_import



"""REST definitions for the resource entity of the api service."""

from waylay.service import WaylayResource
from waylay.service import decorators

DEFAULT_DECORATORS = [decorators.exception_decorator, decorators.return_body_decorator]


class ResourceResource(WaylayResource):
    """REST Resource for the `resource` entity of the `api` (resource provisioning) service."""

    link_roots = {
        'doc': '${doc_url}/api/resources/#',
        'iodoc': '${iodoc_url}/api/resources/?id='
    }

    actions = {
        'get': {
            'method': 'GET',
            'url': '/api/resources/{}',
            'decorators': DEFAULT_DECORATORS,
            'description': 'Retrieve a `resource` representation.',
            'links': {
                'doc': 'retrieve-resource',
                'iodoc': 'retrieve-resource'
            },
        },
        'create': {
            'method': 'POST',
            'url': '/api/resources',
            'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator(['entity'])
            ],
            'description': 'Create a `resource` entity.',
            'links': {
                'doc': 'create-resource',
                'iodoc': 'create-resource'
            },
        },
        'update': {
            'method': 'PATCH',
            'url': '/api/resources/{}',
            'decorators': DEFAULT_DECORATORS,
            'description': '(Partially) update a `resource` representation.',
            'links': {
                'doc': 'partial-resource-update',
                'iodoc': 'partial-resource-update'
            },
        },
        'replace': {
            'method': 'PUT',
            'url': '/api/resources/{}',
            'decorators': DEFAULT_DECORATORS,
            'description': 'Replace a `resource` representation.',
            'links': {
                'doc': 'update-resource',
                'iodoc': 'update-resource'
            },
        },
        'remove': {
            'method': 'DELETE',
            'url': '/api/resources/{}',
            'decorators': DEFAULT_DECORATORS,
            'description': 'Delete a `resource` entity.',
            'links': {
                'doc': 'delete-resource',
                'iodoc': 'delete-resource'
            },
        },
        'search': {
            'method': 'GET',
            'url': '/api/resources',
            'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator(['values'])
            ],
            'description': 'Query `resource` entities.',
            'links': {
                'doc': 'query-resources',
                'iodoc': 'query-resources'
            },
        },
    }




"""REST client for the Waylay API Service (engine)."""

from typing import Optional
from waylay.service import WaylayService

from .resource import ResourceResource
from .resource_type import ResourceTypeResource
from .settings import SettingsResource


class ApiService(WaylayService):
    """REST client for the main Waylay Api Service (engine)."""

    config_key = 'api'
    service_key = 'api'
    default_root_url: Optional[str] = None  # will default to https://{token.domain}
    resource_definitions = {
        'settings': SettingsResource,
        'resource': ResourceResource,
        'resource_type': ResourceTypeResource
    }
    settings: SettingsResource
    resource: ResourceResource
    resource_type: ResourceTypeResource



"""waylay-beta build configuration"""
from setuptools import setup, find_namespace_packages
import versioneer

with open("doc/dist.README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='waylay-beta',
    description='beta release of the Waylay Python SDK',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    url='https://docs-io.waylay.io/#/api/sdk/python',
    author='Waylay',
    author_email='info@waylay.io',
    license='ISC',
    license_file='LICENSE.txt',
    packages=find_namespace_packages(),
    package_data={"waylay": ["py.typed"]},
    include_package_data=True,
    install_requires=[
        'httpx',
        'simple-rest-client',
        'appdirs',
        'python-jose',
        'pandas',
        'isodate',
        'joblib',
        'tqdm',  # progres bar
        'tenacity'
    ],
    extras_require={
        'dev': [
            'pytest',
            'pytest-cov',
            'pytest-mock',
            'mock',
            'pylint',
            'pycodestyle',
            'pydocstyle',
            'autopep8',
            'mypy',
            'typing-inspect',
            'pdoc',
        ],
        'sklearn': [
            'scikit-learn==0.22',
            'dill'
        ],
        'tensorflow': [
            'tensorflow==2.1.0'
        ],
        'pytorch': [
            'torch==1.4'
        ],
        'xgboost': [
            'xgboost==1.0'
        ],
        ':python_version == "3.6"': [
            'python-dateutil',
            'dataclasses',
            'typing_extensions'
        ],
        ':python_version == "3.7"': [
            'typing_extensions'
        ],
    },
    setup_requires=[
        'setuptools-pep8'
    ],
    python_requires='>=3.6',
    entry_points={
        "console_scripts": [
            "waylaycli = waylay.cli.waylaycli:main"
        ],
        "waylay_services": [
            "beta = waylay.service:SERVICES"
        ]
    }
)



from collections import OrderedDict
from collections.abc import Iterable
from typing import Union, TypeVar, Mapping, Tuple
import dataclasses

import pandas as pd
import numpy as np
import datetime
from pandas.core.base import PandasObject

from .parse_util import (
    to_iso_freq as tsa_to_iso_freq
)

RENDER_OPTION_KEYS = [
    RENDER_OPTION_ROLL_UP, RENDER_OPTION_HIERARCHICAL, RENDER_OPTION_VALUE_KEY,
    RENDER_OPTION_SHOW_LEVELS, RENDER_OPTION_ISO_TIMESTAMP,
    RENDER_OPTION_ROW_KEY, RENDER_OPTION_COLUMN_KEY, RENDER_OPTION_HEADER_ARRAY,
    RENDER_OPTION_DATA_AXIS,
    RENDER_OPTION_KEY_SEPERATOR, RENDER_OPTION_KEY_SKIP_EMPTY,
    RENDER_OPTION_MODE,
    RENDER_OPTION_INCLUDE_WINDOW_SPEC
] = [
    'roll_up', 'hierarchical', 'value_key', 'show_levels', 'iso_timestamp',
    'row_key', 'column_key', 'header_array', 'data_axis', 'key_separator', 'key_skip_empty',
    'mode', 'include_window_spec'
]
HEADER_ARRAY_OPTIONS = [HEADER_ARRAY_OPTION_ROW, HEADER_ARRAY_OPTION_COLUMN] = ['row', 'column']
DATA_AXIS_OPTIONS = [DATA_AXIS_ROW, DATA_AXIS_COLUMN] = ['row', 'column']

RENDER_MODE_NAMES = [
    RENDER_MODE_HEADER_ROW, RENDER_MODE_COMPACT, RENDER_MODE_SERIES, RENDER_MODE_HEADER_COLUMN,
    RENDER_MODE_FLAT_DICT, RENDER_MODE_HIER_DICT, RENDER_MODE_METRIC_FLAT_DICT,
    RENDER_MODE_DESC, RENDER_MODE_UPLOAD, RENDER_MODE_COMPACT_WS, RENDER_MODE_CSV
] = [
    'RENDER_MODE_HEADER_ROW', 'RENDER_MODE_COMPACT', 'RENDER_MODE_SERIES', 'RENDER_MODE_HEADER_COLUMN',
    'RENDER_MODE_FLAT_DICT', 'RENDER_MODE_HIER_DICT', 'RENDER_MODE_METRIC_FLAT_DICT',
    'RENDER_MODE_DESC', 'RENDER_MODE_UPLOAD', 'RENDER_MODE_COMPACT_WS', 'RENDER_MODE_CSV'
]

RENDER_OPTIONS = {
    RENDER_OPTION_ROLL_UP: {
        'description': 'move up attributes on rows (or columns) that are the same for \
            all rows (or columns) to a table attribute. \
            Levels enumerated in \'hierarchical\' are excluded.',
        'type': 'boolean'
    },
    RENDER_OPTION_HIERARCHICAL: {
        'description': 'if true, use hierarchical objects to represent multiple row (or column) dimensions, \
            otherwise multi-keys get concatenated with a dot-delimiter. If the value is \
            a list, only these levels are kept as separate levels, while remaining levels get concatenated keys',
        'type': 'array'
    },
    RENDER_OPTION_VALUE_KEY: {
        'description': 'if set, use this key in the value object to report data values',
        'type': 'string'
    },
    RENDER_OPTION_SHOW_LEVELS: {
        'description': 'if set, report the levels used in the data values (either hierarchical or flat)',
        'type': 'boolean'
    },
    RENDER_OPTION_ISO_TIMESTAMP: {
        'description': 'if set, render timestamps in a row or column index with both epoch and iso representations',
        'type': 'boolean'
    },
    RENDER_OPTION_ROW_KEY: {
        'description': 'if set, use this key as name of the row-dimension for single-dimensional rows',
        'type': 'string'
    },
    RENDER_OPTION_COLUMN_KEY: {
        'description': 'if set, use this key as name of the column-dimension for single-dimensional columns',
        'type': 'string'
    },
    RENDER_OPTION_HEADER_ARRAY: {
        'description': 'if set, report data as an header and an array.  ',
        'enum': HEADER_ARRAY_OPTIONS,
        'default': False
    },
    RENDER_OPTION_DATA_AXIS: {
        'description': 'orientation of the tabular data as a array of arrays',
        'enum': DATA_AXIS_OPTIONS
    },
    RENDER_OPTION_KEY_SEPERATOR: {
        'description': 'character used to concatenate multi-key columns or rows when required',
        'type': 'string',
        'default': '.'
    },
    RENDER_OPTION_KEY_SKIP_EMPTY: {
        'description': 'skip empty values in concatenating multi-key column or row headers',
        'type': 'boolean',
        'default': False
    },
    RENDER_OPTION_MODE: {
        'description': 'if set, use a preconfigured render mode with this name',
        'type': 'string',
        'enum': RENDER_MODE_NAMES
    },
    RENDER_OPTION_INCLUDE_WINDOW_SPEC: {
        'description': 'if set, include window specification in render modes that support it',
        'type': 'boolean'
    }
}

RENDER_MODES = {
    RENDER_MODE_HEADER_ROW: {
        "description": "",
        "options": {
            RENDER_OPTION_ISO_TIMESTAMP: True,
            RENDER_OPTION_HEADER_ARRAY: HEADER_ARRAY_OPTION_ROW,
            RENDER_OPTION_ROLL_UP: False,
            RENDER_OPTION_DATA_AXIS: DATA_AXIS_COLUMN
        }
    },
    RENDER_MODE_COMPACT: {
        "description": "",
        "options": {
            RENDER_OPTION_ISO_TIMESTAMP: False,
            RENDER_OPTION_HEADER_ARRAY: HEADER_ARRAY_OPTION_ROW,
            RENDER_OPTION_ROLL_UP: False,
            RENDER_OPTION_DATA_AXIS: DATA_AXIS_COLUMN
        }
    },
    RENDER_MODE_COMPACT_WS: {
        "description": "",
        "options": {
            RENDER_OPTION_ISO_TIMESTAMP: False,
            RENDER_OPTION_HEADER_ARRAY: HEADER_ARRAY_OPTION_ROW,
            RENDER_OPTION_ROLL_UP: False,
            RENDER_OPTION_DATA_AXIS: DATA_AXIS_COLUMN,
            RENDER_OPTION_INCLUDE_WINDOW_SPEC: True
        }
    },
    RENDER_MODE_SERIES: {
        "description": "report columns seperately and data as array by column",
        "options": {
            RENDER_OPTION_ISO_TIMESTAMP: False,
            RENDER_OPTION_HEADER_ARRAY: HEADER_ARRAY_OPTION_ROW,
            RENDER_OPTION_DATA_AXIS: DATA_AXIS_ROW,
            RENDER_OPTION_ROLL_UP: True,
            RENDER_OPTION_INCLUDE_WINDOW_SPEC: True
        }
    },
    RENDER_MODE_HEADER_COLUMN: {
        "description": "",
        "options": {
            RENDER_OPTION_ISO_TIMESTAMP: True,
            RENDER_OPTION_HEADER_ARRAY: HEADER_ARRAY_OPTION_COLUMN,
            RENDER_OPTION_ROLL_UP: False,
            RENDER_OPTION_DATA_AXIS: DATA_AXIS_ROW
        },
    },
    RENDER_MODE_FLAT_DICT: {
        "description": "",
        "options": {
            RENDER_OPTION_ISO_TIMESTAMP: True,
            RENDER_OPTION_HIERARCHICAL: False,
            RENDER_OPTION_SHOW_LEVELS: True,
            RENDER_OPTION_ROLL_UP: False
        },
    },
    RENDER_MODE_HIER_DICT: {
        "description": "",
        "options": {
            RENDER_OPTION_ISO_TIMESTAMP: True,
            RENDER_OPTION_HIERARCHICAL: True,
            RENDER_OPTION_SHOW_LEVELS: True,
            RENDER_OPTION_ROLL_UP: True
        },
    },
    RENDER_MODE_METRIC_FLAT_DICT: {
        "description": "",
        "options": {
            RENDER_OPTION_ISO_TIMESTAMP: True,
            RENDER_OPTION_HIERARCHICAL: ['metric'],
            RENDER_OPTION_SHOW_LEVELS: False,
            RENDER_OPTION_ROLL_UP: True,
            RENDER_OPTION_KEY_SKIP_EMPTY: True
        },
    },
    RENDER_MODE_DESC: {
        "description": "",
        "options": {
            RENDER_OPTION_ISO_TIMESTAMP: True,
            RENDER_OPTION_HIERARCHICAL: ['metric'],
            RENDER_OPTION_SHOW_LEVELS: True,
            RENDER_OPTION_ROLL_UP: True
        },
    },
    RENDER_MODE_UPLOAD: {
        "description": "",
        "options": {
            RENDER_OPTION_ISO_TIMESTAMP: False,
            RENDER_OPTION_HIERARCHICAL: False,
            RENDER_OPTION_SHOW_LEVELS: False,
            RENDER_OPTION_ROLL_UP: True
        },
    },
}


def get_column_names(data, cfg, sep='/'):
    """Retrieve columnn names for CSV rendering."""
    columns_order = ['resource', 'metric', 'aggregation', 'role']
    columns = []
    column_names = data.columns.names
    for i, column in enumerate(data.columns.values):
        name = cfg.data_options_list[i].get('name')
        if not name:
            # sort by column name using columns_order index
            column_dict = dict(zip(column_names, column))
            sorted_column = [column_dict[column] for column in columns_order if column_dict.get(column)]
            name = sep.join(sorted_column)
        columns.append(name)

    return columns


def df_to_csv(dataframe, cfg):
    """Render a dataframe to a CSV string."""
    dataframe = _prepare_data_frame(dataframe, hierarchical=['resource', 'metric'])
    columns = get_column_names(dataframe, cfg)
    # Use ISO formatting with microsecond support
    return dataframe.to_csv(header=columns, index_label='timestamp', date_format='%Y-%m-%dT%H:%M:%S.%fZ')


def df(pandas_object, render_mode=RENDER_MODE_METRIC_FLAT_DICT, render_options=None):
    """Render a dataframe to a json-serializable value."""
    if (render_options is not None and 'render_mode' in render_options):
        render_mode = render_options['render_mode']
    r_args = RENDER_MODES.get(render_mode, {}).get('options', {}).copy()
    if render_options:
        r_args.update(render_options)
    ha_opt = r_args.get(RENDER_OPTION_HEADER_ARRAY, False)
    if not ha_opt:
        raise NotImplementedError(f'Render in object format not supported')
    if ha_opt == HEADER_ARRAY_OPTION_COLUMN:
        return _df_header_column_array(pandas_object, **r_args)
    if ha_opt == HEADER_ARRAY_OPTION_ROW:
        return _df_header_row_array(pandas_object, **r_args)
    raise ValueError(f'Cannot render with option {RENDER_OPTION_HEADER_ARRAY}={ha_opt}')


def df1(pandas_object, render_mode=RENDER_MODE_METRIC_FLAT_DICT, render_options=None):
    """Render a dataframe to a json-serializable value."""
    r = df(pandas_object, render_mode, render_options)
    if isinstance(r, list):
        return r[0] if len(r) == 1 else {} if len(r) == 0 else r
    return r


def parse_df(data, render_mode=None, render_options=None):
    """Parse a json-compatible data structure to a Dataframe."""
    if render_mode is None:
        if isinstance(data, dict) and 'data' in data:
            if 'rows' in data:
                render_mode = RENDER_MODE_HEADER_COLUMN
            elif 'columns' in data:
                render_mode = RENDER_MODE_HEADER_ROW
        else:
            render_mode = RENDER_MODE_HIER_DICT
    r_args = RENDER_MODES.get(render_mode, {}).get('options', {}).copy()
    if render_options:
        r_args.update(render_options)
    ha = r_args.get(RENDER_OPTION_HEADER_ARRAY, False)
    if not ha:
        raise NotImplementedError(f'Render in object format not supported')
    if ha == HEADER_ARRAY_OPTION_COLUMN:
        return _parse_df_header_column_array(data, **r_args)
    if ha == HEADER_ARRAY_OPTION_ROW:
        return _parse_df_header_row_array(data, **r_args)
    raise ValueError(f'Cannot parse with option {RENDER_OPTION_HEADER_ARRAY}={ha}')


def _parse_df_header_column_array(data, data_axis=DATA_AXIS_ROW, **r_arg):
    if 'rows' not in data:
        raise ValueError('cannot parse row-based data without `rows` specification')
    raise NotImplementedError()


def _parse_df_header_row_array(data, data_axis=DATA_AXIS_COLUMN, **r_arg):

    data_values = data.get('data', [])

    if data_axis == DATA_AXIS_COLUMN:
        dataframe = pd.DataFrame(data_values)
    else:
        dataframe = pd.DataFrame()
        for col_idx, data_series in enumerate(data_values):
            dataframe[col_idx] = data_series

    if 'columns' in data:
        row_names, column_index = _parse_row_names_and_column_index(data['columns'])
        if len(dataframe) == 0:
            dataframe = pd.DataFrame(columns=column_index, index=pd.DatetimeIndex([], name='timestamp'))
        else:
            if row_names[0] == 'timestamp':
                idx = pd.DatetimeIndex(
                    pd.to_datetime(dataframe.iloc[:, 0], unit='ms', utc=True),
                    name='timestamp'
                )
                dataframe = dataframe.iloc[:, len(row_names):].set_index(idx)
                row_names = row_names[1:]
            elif row_names:
                dataframe = dataframe.set_index(list(range(len(row_names))))  # pylint: disable=no-member
                dataframe.index.names = row_names
            dataframe.columns = column_index

    if 'attributes' in data:
        dataframe.tsa.table_attributes.update(data['attributes'])
    return dataframe


def _parse_row_names_and_column_index(column_data):
    row_names = []
    column_names = OrderedDict()
    column_keys = []

    col_ordinal_mapping = {'resource': 0, 'metric': 1, 'aggregate': 2}

    def col_sorting_key(col_item):
        col_key = col_item[0]
        return (col_ordinal_mapping.get(col_key, len(col_ordinal_mapping)), col_key)

    for c in column_data:
        if isinstance(c, str):
            row_names.append(c)
        elif isinstance(c, dict):
            for k, v in sorted(c.items(), key=col_sorting_key):
                column_names[k] = True
    for c in column_data:
        if not isinstance(c, str):
            column_keys.append(tuple(c.get(n, None) for n in column_names.keys()))
    return row_names, pd.MultiIndex.from_tuples(column_keys, names=column_names)


def _keys_and_key_series(index, key, iso_timestamp, **render_options):
    keys = []
    # pylint: disable=function-redefined
    key_series = []
    if isinstance(index, pd.DatetimeIndex):
        key = index.name or 'timestamp'
        keys.append(key)
        ts_values = list(
            (index - pd.Timestamp('1970-01-01T00:00Z')) // pd.Timedelta('1ms')
        ) if len(index) else []
        key_series.append(ts_values)
        if iso_timestamp:
            keys.append(key+'_iso')
            iso_ts_values = list(index.strftime('%Y-%m-%dT%H:%M:%S,%f%z'))
            key_series.append(iso_ts_values)
    elif isinstance(index, pd.MultiIndex):
        for lvl_idx, lvl_name in enumerate(index.names):
            keys.append(lvl_name)
            key_series.append(list(index.levels[lvl_idx]))
    elif isinstance(index, pd.RangeIndex):
        pass
    else:
        keys = [index.name or key]
        key_series.append([
            _render_value(v, **render_options) for v in index
        ])
    return keys, key_series


def _keys_and_key_renderer(index, key, iso_timestamp, **render_options):
    keys = None
    # pylint: disable=function-redefined
    key_renderer = None
    if isinstance(index, pd.DatetimeIndex):
        key = index.name or 'timestamp'
        if iso_timestamp:
            keys = [key, key+'_iso']

            def key_renderer(ts):
                return [int(ts.timestamp() * 1000), ts.isoformat()]
        else:
            keys = [key]

            def key_renderer(ts):
                return [int(ts.timestamp() * 1000)]
    elif isinstance(index, pd.MultiIndex):
        keys = list(index.names)

        def key_renderer(idx):
            return list(idx)
    elif isinstance(index, pd.RangeIndex):
        keys = []

        def key_renderer(idx):
            return []
    else:
        keys = [index.name or key]

        def key_renderer(idx):
            return [_render_value(idx, **render_options)]
    return keys, key_renderer


def _header_renderer(index, key, iso_timestamp):
    # pylint: disable=function-redefined
    header_renderer = None
    if isinstance(index, pd.DatetimeIndex):
        key = index.name or 'timestamp'
        if iso_timestamp:
            def header_renderer(ts):
                return {key: int(round(ts.timestamp() * 1000)), key+'_iso': ts.isoformat()}
        else:
            def header_renderer(ts):
                return {key: int(round(ts.timestamp() * 1000))}
    elif isinstance(index, pd.MultiIndex):
        def header_renderer(col):
            return {index.names[i]: _render_value(v) for i, v in enumerate(col) if v is not None and v != ''}
    elif key or index.name:
        key = index.name or key

        def header_renderer(col):
            return {key: _render_value(col)}
    elif isinstance(index, pd.RangeIndex):
        header_renderer = _render_value
    else:
        header_renderer = _render_value
    return header_renderer


def _df_header_row_array(
        pandas_object, row_key='row',
        column_key=None, iso_timestamp=True, data_axis=DATA_AXIS_COLUMN,
        **render_options
):
    if pandas_object is None:
        pandas_object = pd.DataFrame()
    data_frame = _prepare_data_frame(pandas_object, **render_options)
    column_renderer = _header_renderer(data_frame.columns, column_key, iso_timestamp)
    if data_axis == DATA_AXIS_COLUMN:
        row_keys, row_key_renderer = _keys_and_key_renderer(data_frame.index, row_key, iso_timestamp)
        data = [
            row_key_renderer(k) + [_render_value(v, **render_options) for v in r.values]
            for k, r in data_frame.iterrows()
        ]
    else:
        row_keys, index_series = _keys_and_key_series(data_frame.index, row_key, iso_timestamp)
        data = index_series + [
            [_render_value(v, **render_options) for v in data_frame[col]]
            for col in data_frame.columns
        ]
    columns = (row_keys + [column_renderer(col) for col in data_frame.columns])
    return _add_attributes_and_window_spec(data_frame, OrderedDict(
        columns=columns,
        data=data,
        data_axis=data_axis
    ), **render_options)


def _add_attributes_and_window_spec(data_frame, result, include_window_spec=False, **render_options):
    attributes = data_frame.tsa.table_attributes
    if attributes:
        result['attributes'] = _render_dict(attributes, **render_options)
    if include_window_spec:
        window_spec = data_frame.tsa.window_spec
        if window_spec:
            result['window_spec'] = _render_dict(window_spec, **render_options)
    return result


def _df_header_column_array(
        pandas_object, row_key=None,
        column_key='column', iso_timestamp=True,  data_axis=DATA_AXIS_ROW,
        **render_options
):
    if pandas_object is None:
        pandas_object = pd.DataFrame()
    data_frame = _prepare_data_frame(pandas_object, **render_options)
    row_renderer = _header_renderer(data_frame.index, row_key, iso_timestamp)
    if data_axis == DATA_AXIS_ROW:
        column_keys, column_key_renderer = _keys_and_key_renderer(data_frame.columns, column_key, iso_timestamp)
        data = [
            column_key_renderer(col) + [_render_value(v, **render_options) for v in data_frame[col].values]
            for col in data_frame.columns
        ]
    else:
        column_keys, columns_series = _keys_and_key_series(data_frame.columns, column_key, iso_timestamp)
        data = columns_series + [
            [_render_value(v, **render_options) for v in row_series]
            for idx, row_series in data_frame.iterrows()
        ]
    return _add_attributes_and_window_spec(data_frame, OrderedDict(
        rows=(column_keys + [row_renderer(idx) for idx in data_frame.index]),
        data=data,
        data_axis=data_axis,
    ), **render_options)


def _roll_up_single_level_columns(data_frame, keep_hashed, value_key):
    return _roll_up_single_level(data_frame, 1, keep_hashed, value_key)


def _roll_up_single_level_rows(data_frame, keep_hashed, value_key):
    return _roll_up_single_level(data_frame, 0, keep_hashed, value_key)


def _roll_up_single_level(data_frame, axis, keep_hashed, value_key):
    table_attributes = data_frame.tsa.table_attributes.copy()
    index = data_frame.axes[axis]
    hashed = [name for name in index.names if name in keep_hashed]
    current_level = 0
    while isinstance(index, pd.MultiIndex) and len(index.names) > len(hashed):
        level_name = index.names[current_level]
        labels = set(index.get_level_values(current_level))
        non_empty_labels = [la for la in labels if la != '']
        if level_name in hashed:
            current_level += 1
        elif len(non_empty_labels) == 1:
            # other column dimensions (resource, aggregation, ... )
            # are rolled up if they have only one value (not counting empty labels)
            label = non_empty_labels[0]
            table_attributes[level_name] = label
            rolled_up_data_frame = data_frame.xs(label, axis=axis, level=current_level)
            if len(labels) == 2:
                data_frame = rolled_up_data_frame.join(data_frame.xs(
                    '', axis=axis, level=current_level), lsuffix='_default')
            else:
                data_frame = rolled_up_data_frame
            index = data_frame.axes[axis]
        else:
            hashed = hashed + [level_name]
            current_level += 1
    if len(index.names) == 1 and len(index) == 1 and index.name and index.name not in hashed and value_key:
        table_attributes[index.name] = index[0]
        data_frame = data_frame.set_axis(pd.Index([value_key]), axis=axis, inplace=False)

    data_frame.tsa.table_attributes.update(table_attributes)
    return data_frame


def _prepare_data_frame(
        pandas_object, roll_up=True,
        hierarchical=['metric'], value_key='value', **kwargs
):
    if isinstance(pandas_object, pd.Series):
        if isinstance(pandas_object.index, (pd.DatetimeIndex, pd.RangeIndex)):
            data_frame = pandas_object.to_frame()
        else:
            data_frame = pandas_object.to_frame().T
        if len(data_frame.columns) > 1:
            data_frame.columns.names = ["col"+str(i) for i in range(0, len(data_frame.columns.names))]
    elif isinstance(pandas_object, pd.DataFrame):
        data_frame = pandas_object
    else:
        raise ValueError('not a pandas object: {}'.format(pandas_object))
    if roll_up:
        keep_hashed = hierarchical if isinstance(hierarchical, list) else []
        data_frame = _roll_up_single_level_columns(data_frame, keep_hashed, value_key)
        data_frame = _roll_up_single_level_rows(data_frame, keep_hashed, value_key)
    return data_frame


def _render_multi_index_row_to_dict(row):
    if len(row.index.names) > 1:
        row_dict = {}
        for sub_key, sub_row in row.groupby(level=0):
            if sub_key != '':
                row_dict[sub_key] = _render_multi_index_row_to_dict(sub_row.xs(sub_key))
            else:
                # empty keys get render on level above .. ???
                row_dict.update(_render_multi_index_row_to_dict(sub_row.xs(sub_key)))
        return row_dict
    return row.to_dict()


def _render_key(k, key_separator='.', key_skip_empty=False, **kwargs):
    if isinstance(k, str):
        return k
    if isinstance(k, Iterable):
        if key_skip_empty:
            return key_separator.join([str(v) for v in k if v != ''])
        return key_separator.join([str(v) for v in k])
    return str(k)


def render(value, **render_options):
    """Convert a value to a json-serializable one."""
    return _render_value(value, **render_options)


def render_timestamp(value, iso_timestamp=False, **render_options):
    """Render a timestamp."""
    if iso_timestamp:
        return value.isoformat()
    return int(value.timestamp() * 1000)


def _render_value(v, **render_options):
    if isinstance(v, str):
        return v
    if isinstance(v, pd.Timestamp):
        return render_timestamp(v, **render_options)
    if isinstance(v, datetime.datetime):
        return render_timestamp(v, True, **render_options)
    if isinstance(v, pd.Timedelta):
        return tsa_to_iso_freq(v)
    if isinstance(v, np.generic):
        if np.isfinite(v):
            return v.item()
        return None
    if isinstance(v, float):
        if np.isfinite(v):
            return v
        return None
    if isinstance(v, list):
        return [_render_value(item, **render_options) for item in v]
    if isinstance(v, dict):
        return _render_dict(v, **render_options)
    if isinstance(v, np.ndarray):
        return [_render_value(item, **render_options) for item in v]
    if hasattr(v, 'render'):
        return v.render()
    if dataclasses.is_dataclass(v):
        return OrderedDict([
            (f.name, _render_value(getattr(v, f.name), **render_options))
            for f in dataclasses.fields(v)
        ])
    if isinstance(v, PandasObject):
        return df1(v, render_options=render_options)
    if pd.isnull(v):
        return None
    return v


def _render_dict(dict_like, **render_options):
    if len(dict_like) == 1 and '' in dict_like:
        return _render_value(dict_like[''], **render_options)

    return OrderedDict([
        (_render_key(key, **render_options), _render_value(val, **render_options)) for key, val in dict_like.items()
    ])



"""REST definitions for the 'bucket' entity of the 'storage' service."""
from urllib.parse import quote

from waylay.service import WaylayResource, decorators


class BucketResource(WaylayResource):
    """REST Resource for the 'bucket' entity of the 'storage' service."""

    link_roots = {
        'iodoc': '${iodoc_url}/api/storage/?id=',
        'apidoc':  '${root_url}/docs#/storage/',
        'openapi': '${root_url}/openapi.json'
    }

    actions = {
        'list': {
            'method': 'GET', 'url': '/bucket', 'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator(['buckets']),
            ],
            'description': 'List available bucket aliases',
            'links': {
                'iodoc': 'list-bucket',
                'apidoc': 'list_buckets_bucket__get',
                'openapi': f"#/paths/{quote('/bucket')}/get"
            }
        },
        'get': {
            'method': 'GET', 'url': '/bucket/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_body_decorator,
            ],
            'description': 'Get metadata for a specific bucket alias',
            'links': {
                'iodoc': 'get-bucket',
                'apidoc': 'get_bucket_bucket__bucket_name__get',
                'openapi': f"#/paths/{quote('/bucket/{bucket_name}')}/get"
            }
        },
    }



"""Reader utilities for a etl-formatted file."""
from typing import (
    Iterator, Sequence, List,
    Union, Optional, Callable,
    Tuple, Any
)

from datetime import datetime
import csv
import gzip
from contextlib import contextmanager

import pandas as pd

from .model import (
    WaylayETLSeriesImport,
    SeriesIterator,
    MeasurementValue,
    MeasurementIterator,
    Resource,
    Metric,
    SeriesSettings,
    CSVReader,
    METRIC_PREFIX,
    METRIC_PREFIX_LENGTH,
    RESOURCE_IDX, METRIC_IDX, TIMESTAMP_IDX, VALUE_IDX,
    RESOURCE_COLUMN, METRIC_COLUMN, TIMESTAMP_COLUMN,
    ParserRequestError,
    try_parse_float,
    parse_timestamp,
)


@contextmanager
def read_etl_import_as_stream(etl_import: WaylayETLSeriesImport) -> Iterator[CSVReader]:
    """Iterate of the content of stored etl file (gzipped csv file)."""
    path = etl_import.import_file.path
    if not path.exists():
        raise ParserRequestError(
            f'Cannot read from etl import file {path}: does not exist.'
        )
    with gzip.open(path, 'rt') as csv_file:
        yield csv.reader(csv_file, delimiter=',', quoting=csv.QUOTE_MINIMAL)


@contextmanager
def read_etl_import(etl_import: WaylayETLSeriesImport) -> Iterator[SeriesIterator]:
    """Create a Series Iterator from a stored etl file (gzipped csv file)."""
    path = etl_import.import_file.path
    if not path.exists():
        raise ParserRequestError(
            f'Cannot read from etl import file {path}: does not exist.'
        )
    with gzip.open(path, 'rt') as csv_file:
        reader = csv.reader(csv_file, delimiter=',', quoting=csv.QUOTE_MINIMAL)
        yield SeriesCursor(reader, list(etl_import.settings.iter_metrics())).iter_series()


def dataframe_from_iterator(series_iterator: SeriesIterator) -> pd.DataFrame:
    """Create a Pandas dataframe from a SeriesIterator."""
    resource_metric_dfs = list(
        pd.DataFrame(
            list(measurements),
            columns=[TIMESTAMP_COLUMN, (resource, metric)]
        ).set_index(TIMESTAMP_COLUMN)
        for resource, metric, measurements in series_iterator
    )
    resource_metric_df = pd.concat(resource_metric_dfs, axis=1)
    resource_metric_df.columns = pd.MultiIndex.from_tuples(
        resource_metric_df.columns,
        names=[RESOURCE_COLUMN, METRIC_COLUMN]
    )
    return resource_metric_df


def list_resources(settings: SeriesSettings, series_iterator: SeriesIterator) -> List[Resource]:
    """Get a `Resource` list that describes the resources an metrics found in the given series collection.

    Uses the `settings.resources` and `settings.metrics` as a data dictionary to enhance the
    _resource_ and _metric_ descriptions found in the series.
    """
    class ResourceDict(dict):
        def __missing__(self, key):
            self[key] = value = settings.resource_for(key, [])
            return value

    resource_dict = ResourceDict()
    for resource, metric, _ in series_iterator:
        resource_dict[resource].metrics.append(settings.metric_for(metric))

    return list(resource_dict.values())


DONE = False
NEXT = True

RESOURCE_IDX, METRIC_IDX, TIMESTAMP_IDX, VALUE_IDX = range(4)
SeriesRow = Tuple[Any]


class SeriesCursor:
    """A look-ahead polling cursor on a etl-formatted csv stream, splitting the series."""

    def __init__(
        self, reader,
        metrics: Optional[Sequence[Metric]] = None,
        on_close: Optional[Callable] = None
    ):
        """Create a Series reader.

        The reader should provide an iterator of string-formatted
        (timestamp, resource, metric, value) tuples.
        """
        self.reader = reader
        self._current_row: Union[None, bool, SeriesRow] = None
        self._next_row: Union[None, bool, SeriesRow] = None
        self._on_close: Optional[Callable] = on_close
        self._metrics: Optional[Sequence[Metric]] = metrics
        self._value_parser: Callable[[str], MeasurementValue] = try_parse_float
        self._metric_prefix_len = 0

    def _current_value_parser(self) -> Callable:
        # look up the value parser for the metric at the current cursor position.
        metric = self.metric
        if self._metrics:
            for metric_spec in self._metrics:
                if metric_spec.name == metric:
                    return metric_spec.get_value_parser()
        return try_parse_float

    def iter_rows(self) -> MeasurementIterator:
        """Get an iterator of the series at the current cursor position.

        Precondition:
        - a previous call to `next_series` should have returned `True`.
        """
        while self.next_row():
            yield self.timestamp, self.value

    def iter_series(self) -> SeriesIterator:
        """Get an iterator over (_metric_, _resource_, _series_) tuples.

        Warning:
            Each _series_ iterator returned in the tuple
            should first be depleted before handling the next record of this iterator.
            Out of order requests will lead to lost rows.
        """
        while self.next_series():
            yield self.resource, self.metric, self.iter_rows()  # type: ignore

    def next_series(self) -> bool:
        """Move the cursor to the next series.

        If this returns `False`, the cursor is at the end of the input.
        If this returns `True`, the `resource` and `metric` properties can be requested,
        and the available series can be iterated with `next_series`.
        """
        if self._current_row == NEXT:
            # consume all rows for this series
            while self.next_row():
                pass

        self._current_row = NEXT

        if self._next_row is None:
            # read first row
            self._next_row = next(self.reader, DONE)

        if self._next_row != DONE and self._next_row[TIMESTAMP_IDX] == TIMESTAMP_COLUMN:  # type: ignore
            # consume header row
            self._next_row = next(self.reader, DONE)

        if self._next_row == DONE:
            # last entry
            if self._on_close:
                self._on_close()
            return False

        # setup specific parser rules for this series.
        self._value_parser = self._current_value_parser()

        # inspect the prefixes of metrics, set up to ignore METRIC_PREFIX
        has_metric_prefix = METRIC_PREFIX in self._next_row[METRIC_IDX]  # type: ignore
        self._metric_prefix_len = METRIC_PREFIX_LENGTH if has_metric_prefix else 0

        return True

    def next_row(self) -> bool:
        """Move the cursor to the next row.

        Only if this returns `true`, the `value` and `timestamp` properties can be requested.
        If it returns `false`, a call to `next_series` should be issued to check completion.

        Precondition:
        - a previous call to `next_series` should have returned `True`.
        """
        if self._next_row is None:
            raise AssertionError('Illegal state: call `next_series` first!')

        if self._next_row == DONE:
            # last entry
            return False

        if self._current_row != NEXT and (
            self._next_row[RESOURCE_IDX] != self._current_row[RESOURCE_IDX] or  # type: ignore
            self._next_row[METRIC_IDX] != self._current_row[METRIC_IDX]    # type: ignore
        ):
            # new series (resource/metric changed)
            return False

        # read next row
        self._current_row = self._next_row
        self._next_row = next(self.reader, DONE)
        return True

    @property
    def resource(self) -> Optional[str]:
        """Get the _resource_ for the current cursor position.

        Precondition:
        - a previous call to `next_series` should have returned `True`.
        """
        return self._current_or_next(RESOURCE_IDX)

    @property
    def metric(self) -> Optional[str]:
        """Get the _metric_ for the current cursor position.

        Precondition:
        - a previous call to `next_series` should have returned `True`.
        """
        _metric = self._current_or_next(METRIC_IDX)
        return _metric[self._metric_prefix_len:]

    @property
    def timestamp(self) -> datetime:
        """Get the _timestamp_ for the current cursor position.

        Precondition:
        - a previous call to `next_row` should have returned `True`.
        """
        return parse_timestamp(self._current_row[TIMESTAMP_IDX])  # type: ignore

    @property
    def value(self) -> MeasurementValue:
        """Get the _value_ for the current cursor position.

        Precondition:
        - a previous call to `next_row` should have returned `True`.
        """
        return self._value_parser(self._current_row[VALUE_IDX])  # type: ignore

    def _current_or_next(self, idx):
        if self._current_row == NEXT:
            return self._next_row[idx]
        return self._current_row[idx]
